@@ -20,12 +20,9 @@
  */
 
 #include "fsmgraph.h"
-
-#ifdef LOG_CONDS
 #include <iostream>
 using std::cerr;
 using std::endl;
-#endif
 
 CondData *condData = 0;
 KeyOps *keyOps = 0;
@@ -237,7 +234,7 @@ void FsmAp::fillGaps( StateAp *state )
 {
 	if ( state->outList.length() == 0 ) {
 		/* Add the range on the lower and upper bound. */
-		attachNewTrans( state, 0, keyOps->lowKey, keyOps->highKey );
+		attachNewTrans( state, 0, keyOps->minKey, keyOps->maxKey );
 	}
 	else {
 		TransList srcList;
@@ -245,12 +242,12 @@ void FsmAp::fillGaps( StateAp *state )
 
 		/* Check for a gap at the beginning. */
 		TransList::Iter trans = srcList, next;
-		if ( keyOps->lt( keyOps->lowKey, trans->lowKey ) ) {
+		if ( keyOps->minKey < trans->lowKey ) {
 			/* Make the high key and append. */
-			long highKey = trans->lowKey;
-			keyOps->dec( highKey );
+			Key highKey = trans->lowKey;
+			highKey.decrement();
 
-			attachNewTrans( state, 0, keyOps->lowKey, highKey );
+			attachNewTrans( state, 0, keyOps->minKey, highKey );
 		}
 
 		/* Write the transition. */
@@ -258,19 +255,19 @@ void FsmAp::fillGaps( StateAp *state )
 		state->outList.append( trans );
 
 		/* Keep the last high end. */
-		long lastHigh = trans->highKey;
+		Key lastHigh = trans->highKey;
 
 		/* Loop each source range. */
 		for ( trans = next; trans.lte(); trans = next ) {
 			/* Make the next key following the last range. */
-			long nextKey = lastHigh;
-			keyOps->inc( nextKey );
+			Key nextKey = lastHigh;
+			nextKey.increment();
 
 			/* Check for a gap from last up to here. */
-			if ( keyOps->lt( nextKey, trans->lowKey ) ) {
+			if ( nextKey < trans->lowKey ) {
 				/* Make the high end of the range that fills the gap. */
-				long highKey = trans->lowKey;
-				keyOps->dec( highKey );
+				Key highKey = trans->lowKey;
+				highKey.decrement();
 
 				attachNewTrans( state, 0, nextKey, highKey );
 			}
@@ -284,11 +281,11 @@ void FsmAp::fillGaps( StateAp *state )
 		}
 
 		/* Now check for a gap on the end to fill. */
-		if ( keyOps->lt( lastHigh, keyOps->highKey ) ) {
+		if ( lastHigh < keyOps->maxKey ) {
 			/* Get a copy of the default. */
-			keyOps->inc( lastHigh );
+			lastHigh.increment();
 
-			attachNewTrans( state, 0, lastHigh, keyOps->highKey );
+			attachNewTrans( state, 0, lastHigh, keyOps->maxKey );
 		}
 	}
 }
@@ -639,9 +636,10 @@ void FsmAp::nullActionKeys( )
 void FsmAp::verifyStates()
 {
 	for ( StateList::Iter state = stateList; state.lte(); state++ ) {
-		/* Non final states should not have leaving priorities or transitions. */
+		/* Non final states should not have leaving data. */
 		if ( ! (state->stateBits & SB_ISFINAL) ) {
 			assert( state->outActionTable.length() == 0 );
+			assert( state->outCondSet.length() == 0 );
 			assert( state->outPriorTable.length() == 0 );
 		}
 
@@ -757,6 +755,12 @@ int FsmAp::compareStateData( const StateAp *state1, const StateAp *state2 )
 	if ( cmpRes != 0 )
 		return cmpRes;
 
+	/* Test out condition sets. */
+	cmpRes = CmpActionSet::compare( state1->outCondSet, 
+			state2->outCondSet );
+	if ( cmpRes != 0 )
+		return cmpRes;
+
 	/* Test out error action tables. */
 	cmpRes = CmpErrActionTable::compare( state1->errActionTable, 
 			state2->errActionTable );
@@ -768,115 +772,69 @@ int FsmAp::compareStateData( const StateAp *state1, const StateAp *state2 )
 			state2->eofActionTable );
 }
 
-/* Callback invoked when destTrans is made into a transition leaving srcState.
- * Draw in the out transition properties of srcState into this transition.
- * LeavingFromState is called when concatOp or starOp is called with
- * leavingFsm = true and a transition is made that represents leaving machine
- * one and going to machine two in the case of concatOp and leaving a machine
- * to go back into itself in the case of the star operator. */
-void FsmAp::leavingFromState( TransAp *destTrans, StateAp *srcState )
-{
-	/* Get the actions data from the outActionTable. */
-	destTrans->actionTable.setActions( srcState->outActionTable );
 
-	/* Get the priorities from the outPriorTable. */
-	destTrans->priorTable.setPriors( srcState->outPriorTable );
-}
-
-
-/* Callback invoked when a state looses its final state status. This is where
- * properties of final states get unset. At the point of invocation, it is
- * unspecified whether or not isFinState is set.  Callback invoked when a
- * state looses its final state status. This is where properties of final
- * states get unset. At the point of invocation, it is unspecified whether or
- * not isFinState is set. */
-void FsmAp::relinquishFinal( StateAp *state )
+/* Invoked when a state looses its final state status and the leaving
+ * transition embedding data should be deleted. */
+void FsmAp::clearOutData( StateAp *state )
 {
 	/* Kill the out actions and priorities. */
 	state->outActionTable.empty();
+	state->outCondSet.empty();
 	state->outPriorTable.empty();
 }
 
+bool FsmAp::hasOutData( StateAp *state )
+{
+	return ( state->outActionTable.length() > 0 ||
+			state->outCondSet.length() > 0 ||
+			state->outPriorTable.length() > 0 );
+}
+
 /* 
  * Setting Conditions.
  */
 
-long FsmAp::baseKey( long key, const CondSet &condSet )
+
+void logNewExpansion( Expansion *exp );
+void logCondSpace( CondSpace *condSpace );
+
+CondSpace *FsmAp::addCondSpace( const CondSet &condSet )
 {
-	Condition *cond = condData->condDict.find( key );
-	assert( cond != 0 );
+	CondSpace *condSpace = condData->condSpaceMap.find( condSet );
+	if ( condSpace == 0 ) {
+		Key baseKey = condData->nextCondKey;
+		condData->nextCondKey += (1 << condSet.length() ) * keyOps->alphSize();
 
-	CondSetMapEl *csMapEl = cond->condSetMap.find( condSet );
-	assert( csMapEl != 0 );
-	return csMapEl->baseKey;
-}
-
-bool FsmAp::findCondition( long key, Condition *&cond, CondSetMapEl *&csMapEl )
-{
-	for ( CondDict::Iter cdi = condData->condDict; cdi.lte(); cdi++ ) {
-		for ( CondSetMap::Iter cs = cdi->condSetMap; cs.lte(); cs++ ) {
-			int highKey = cs->baseKey + ( 1 << cs->condSet.length() );
-			if ( cs->baseKey <= key && key <= highKey ) {
-				cond = cdi;
-				csMapEl = cs;
-				return true;
-			}
-		}
-	}
-	return false;
-}
-
-CondSetMapEl *FsmAp::addCondition( long key, const CondSet &condSet )
-{
-	assert( condSet.length() > 0 );
-
-	Condition *cond = condData->condDict.find( key );
-	if ( cond == 0 ) {
-		#ifdef LOG_CONDS
-		cerr << "adding condition key = " << key << endl;
-		#endif
-		cond = new Condition( key );
-		condData->condDict.insert( cond );
-	}
-
-	CondSetMapEl *csMapEl = cond->condSetMap.find( condSet );
-	if ( csMapEl == 0 ) {
-		int numCondKeys = 1 << condSet.length();
-
-		long lowKey = condData->nextCondKey;
-		condData->nextCondKey += numCondKeys;
+		condSpace = new CondSpace( condSet );
+		condSpace->baseKey = baseKey;
+		condData->condSpaceMap.insert( condSpace );
 
 		#ifdef LOG_CONDS
-		cerr << "adding condition set length = " << condSet.length() << 
-				" numCondKeys = " << numCondKeys << endl;
+		cerr << "adding new condition space" << endl;
+		cerr << "  condition set: ";
+		logCondSpace( condSpace );
+		cerr << endl;
+		cerr << "  baseKey: " << baseKey.getVal() << endl;
 		#endif
-		csMapEl = cond->condSetMap.insert( condSet );
-		csMapEl->baseKey = lowKey;
 	}
-
-	return csMapEl;
+	return condSpace;
 }
 
-void FsmAp::setCondition( Action *condAction )
+void FsmAp::startFsmCondition( Action *condAction )
 {
-	assert( stateList.length() == 2 );
-	StateAp *state = stateList.head;
-
-	assert( state->stateCondDict.length() == 0 );
-	assert( state->outList.length() == 1 );
-	TransAp *trans = state->outList.head;
-
-	assert( trans->lowKey == trans->highKey );
-	assert( state->next->outList.length() == 0 );
-
-	long key = trans->lowKey;
-	CondSet condSet( condAction );
-
-	CondSetMapEl *csMapEl = addCondition( key, condSet );
-
-	StateCond *stateCond = new StateCond( trans->lowKey );
-	stateCond->condSet.insert( condAction );
-
-	state->stateCondDict.insert( stateCond );
-	trans->lowKey = trans->highKey = csMapEl->baseKey+1;
+	/* Make sure the start state has no other entry points. */
+	isolateStartState();
+	embedCondition( startState, condAction );
+}
+
+void FsmAp::allTransCondition( Action *condAction )
+{
+	for ( StateList::Iter state = stateList; state.lte(); state++ )
+		embedCondition( state, condAction );
+}
+
+void FsmAp::leaveFsmCondition( Action *condAction )
+{
+	for ( StateSet::Iter state = finStateSet; state.lte(); state++ )
+		(*state)->outCondSet.insert( condAction );
 }
