@@ -88,6 +88,9 @@ struct ActionTable
 	bool hasAction( Action *action );
 };
 
+typedef SBstSet< Action*, CmpOrd<Action*> > ActionSet;
+typedef CmpSTable< Action*, CmpOrd<Action*> > CmpActionSet;
+
 /* Transistion Action Element. */
 typedef SBstMapEl< int, LongestMatchPart* > LmActionTableEl;
 
@@ -282,9 +285,12 @@ struct TransAp
 		highKey(other.highKey),
 		fromState(0), toState(0),
 		actionTable(other.actionTable),
-		priorTable(other.priorTable) {}
+		priorTable(other.priorTable)
+	{
+		assert( lmActionTable.length() == 0 && other.lmActionTable.length() == 0 );
+	}
 
-	long lowKey, highKey;
+	Key lowKey, highKey;
 	StateAp *fromState;
 	StateAp *toState;
 
@@ -381,24 +387,31 @@ struct TransEl
 {
 	/* Constructors. */
 	TransEl() { }
-	TransEl( long lowKey, long highKey ) 
+	TransEl( Key lowKey, Key highKey ) 
 		: lowKey(lowKey), highKey(highKey) { }
-	TransEl( long lowKey, long highKey, TransAp *value ) 
+	TransEl( Key lowKey, Key highKey, TransAp *value ) 
 		: lowKey(lowKey), highKey(highKey), value(value) { }
 
-	long lowKey, highKey;
+	Key lowKey, highKey;
 	TransAp *value;
 };
 
-/* Vector based set of key items. */
-struct KeySet
-:
-	public Vector<long>
+struct CmpKey
 {
-	/* Return a pointer to the item in the set on success. */
-	long *insert( long key );
+	static int compare( const Key key1, const Key key2 )
+	{
+		if ( key1 < key2 )
+			return -1;
+		else if ( key1 > key2 )
+			return 1;
+		else
+			return 0;
+	}
 };
 
+/* Vector based set of key items. */
+typedef BstSet<Key, CmpKey> KeySet;
+
 struct MinPartition 
 {
 	MinPartition() : active(false) { }
@@ -431,79 +444,84 @@ typedef BstSet<LongestMatchPart*> LmItemSet;
 
 /* Conditions. */
 typedef BstSet< Action*, CmpOrd<Action*> > CondSet;
-typedef Vector< long > LongVect;
-typedef Vector<LongVect*> Expansions;
-
 typedef CmpTable< Action*, CmpOrd<Action*> > CmpCondSet;
 
-struct CondKey
+struct CondSpace
+	: public AvlTreeEl<CondSpace>
 {
-	CondKey( long key, Action *cond )
-		: key(key), cond(cond) {}
-
-	long key;
-	Action *cond;
-};
-
-struct CmpCondKey
-{
-	int compare( long k1, long k2 ) const
-	{
-		if ( keyOps->lt( k1, k2 ) )
-			return -1;
-		else if ( keyOps->gt( k1, k2 ) )
-			return 1;
-		return 0;
-	}
-};
-
-struct CondSetMapEl
-	: public AvlTreeEl<CondSetMapEl>
-{
-	CondSetMapEl( const CondSet &condSet )
-		: condSet(condSet), conditionId(0) {}
+	CondSpace( const CondSet &condSet )
+		: condSet(condSet) {}
 	
 	const CondSet &getKey() { return condSet; }
 
 	CondSet condSet;
-	long baseKey;
-	int conditionId;
+	Key baseKey;
+	long condSpaceId;
 };
 
-typedef AvlTree<CondSetMapEl, CondSet, CmpCondSet> CondSetMap;
+typedef Vector<CondSpace*> CondSpaceVect;
 
-struct Condition
-	: AvlTreeEl<Condition>
-{
-	Condition( long key ) 
-		: key(key) {}
-	long getKey() { return key; }
-
-	long key;
-
-	CondSetMap condSetMap;
-};
-typedef AvlTree<Condition, long, CmpCondKey> CondDict;
+typedef AvlTree<CondSpace, CondSet, CmpCondSet> CondSpaceMap;
 
 struct StateCond
-	: AvlTreeEl<StateCond>
 {
-	StateCond( long key ) : key(key) {}
-	long getKey() { return key; }
+	StateCond( Key lowKey, Key highKey ) :
+		lowKey(lowKey), highKey(highKey) {}
 
-	long key;
+	Key lowKey;
+	Key highKey;
+	CondSpace *condSpace;
 
-	CondSet condSet;
+	StateCond *prev, *next;
+};
+
+typedef DList<StateCond> StateCondList;
+typedef Vector<long> LongVect;
+
+struct Expansion
+{
+	Expansion( Key lowKey, Key highKey ) :
+		lowKey(lowKey), highKey(highKey),
+		fromTrans(0), fromCondSpace(0), 
+		toCondSpace(0) {}
+	
+	~Expansion()
+	{
+		if ( fromTrans != 0 )
+			delete fromTrans;
+	}
+
+	Key lowKey;
+	Key highKey;
+
+	TransAp *fromTrans;
+	CondSpace *fromCondSpace;
+	long fromVals;
+
+	CondSpace *toCondSpace;
+	LongVect toValsList;
+
+	Expansion *prev, *next;
+};
+
+typedef DList<Expansion> ExpansionList;
+
+struct Removal
+{
+	Key lowKey;
+	Key highKey;
+
+	Removal *next;
 };
-typedef AvlTree<StateCond, long, CmpCondKey> StateCondDict;
 
 struct CondData
 {
 	CondData() : nextCondKey(0) {}
 
 	/* Condition info. */
-	CondDict condDict;
-	long nextCondKey;
+	Key nextCondKey;
+
+	CondSpaceMap condSpaceMap;
 };
 
 extern CondData *condData;
@@ -531,7 +549,7 @@ struct StateAp
 	EpsilonTrans epsilonTrans;
 
 	/* Condition info. */
-	StateCondDict stateCondDict;
+	StateCondList stateCondList;
 
 	/* Number of in transitions from states other than ourselves. */
 	int foreignInTrans;
@@ -597,6 +615,9 @@ struct StateAp
 	/* Actions to add to any future transitions that leave via this state. */
 	ActionTable outActionTable;
 
+	/* Conditions to add to any future transiions that leave via this sttate. */
+	ActionSet outCondSet;
+
 	/* Error action tables. */
 	ErrActionTable errActionTable;
 
@@ -607,21 +628,23 @@ struct StateAp
 	LmItemSet lmItemSet;
 };
 
-struct NextTrans
+template <class ListItem> struct NextTrans
 {
-	long lowKey, highKey;
-	TransAp *trans;
-	TransAp *next;
+	Key lowKey, highKey;
+	ListItem *trans;
+	ListItem *next;
 
 	void load() {
-		if ( trans != 0 ) {
+		if ( trans == 0 )
+			next = 0;
+		else {
 			next = trans->next;
 			lowKey = trans->lowKey;
 			highKey = trans->highKey;
 		}
 	}
 
-	void set( TransAp *t ) {
+	void set( ListItem *t ) {
 		trans = t;
 		load();
 	}
@@ -632,7 +655,16 @@ struct NextTrans
 	}
 };
 
-struct PairIter
+
+/* Encodes the different states that are meaningful to the of the iterator. */
+enum PairIterUserState
+{
+	RangeInS1, RangeInS2,
+	RangeOverlap,
+	BreakS1, BreakS2
+};
+
+template <class ListItem1, class ListItem2 = ListItem1> struct PairIter
 {
 	/* Encodes the different states that an fsm iterator can be in. */
 	enum IterState {
@@ -646,15 +678,7 @@ struct PairIter
 		ExactOverlap,   End
 	};
 
-	/* Encodes the different states that are meaningful to the of the
-	 * iterator. */
-	enum UserState {
-		RangeInS1, RangeInS2,
-		RangeOverlap,
-		BreakS1, BreakS2
-	};
-
-	PairIter( const StateAp *state1, const StateAp *state2, bool wantBreaks );
+	PairIter( ListItem1 *list1, ListItem2 *list2 );
 	
 	/* Query iterator. */
 	bool lte() { return itState != End; }
@@ -663,20 +687,222 @@ struct PairIter
 	void operator++()    { findNext(); }
 
 	/* Iterator state. */
-	const StateAp *state1;
-	const StateAp *state2;
+	ListItem1 *list1;
+	ListItem2 *list2;
 	IterState itState;
-	UserState userState;
+	PairIterUserState userState;
 
-	NextTrans s1Tel, s2Tel;
-	long bottomLow, bottomHigh;
-	TransAp *bottomTrans;
-	bool wantBreaks;
+	NextTrans<ListItem1> s1Tel;
+	NextTrans<ListItem2> s2Tel;
+	Key bottomLow, bottomHigh;
+	ListItem1 *bottomTrans1;
+	ListItem2 *bottomTrans2;
 
 private:
 	void findNext();
 };
 
+/* Init the iterator by advancing to the first item. */
+template <class ListItem1, class ListItem2> PairIter<ListItem1, ListItem2>::PairIter( 
+		ListItem1 *list1, ListItem2 *list2 )
+:
+	list1(list1),
+	list2(list2),
+	itState(Begin)
+{
+	findNext();
+}
+
+/* Return and re-entry for the co-routine iterators. This should ALWAYS be
+ * used inside of a block. */
+#define CO_RETURN(label) \
+	itState = label; \
+	return; \
+	entry##label: backIn = true
+
+/* Return and re-entry for the co-routine iterators. This should ALWAYS be
+ * used inside of a block. */
+#define CO_RETURN2(label, uState) \
+	itState = label; \
+	userState = uState; \
+	return; \
+	entry##label: backIn = true
+
+/* Advance to the next transition. When returns, trans points to the next
+ * transition, unless there are no more, in which case end() returns true. */
+template <class ListItem1, class ListItem2> void PairIter<ListItem1, ListItem2>::findNext()
+{
+	/* This variable is used in dummy statements that follow the entry
+	 * goto labels. The compiler needs some statement to follow the label. */
+	bool backIn;
+
+	/* Jump into the iterator routine base on the iterator state. */
+	switch ( itState ) {
+		case Begin:              goto entryBegin;
+		case ConsumeS1Range:     goto entryConsumeS1Range;
+		case ConsumeS2Range:     goto entryConsumeS2Range;
+		case OnlyInS1Range:      goto entryOnlyInS1Range;
+		case OnlyInS2Range:      goto entryOnlyInS2Range;
+		case S1SticksOut:        goto entryS1SticksOut;
+		case S1SticksOutBreak:   goto entryS1SticksOutBreak;
+		case S2SticksOut:        goto entryS2SticksOut;
+		case S2SticksOutBreak:   goto entryS2SticksOutBreak;
+		case S1DragsBehind:      goto entryS1DragsBehind;
+		case S1DragsBehindBreak: goto entryS1DragsBehindBreak;
+		case S2DragsBehind:      goto entryS2DragsBehind;
+		case S2DragsBehindBreak: goto entryS2DragsBehindBreak;
+		case ExactOverlap:       goto entryExactOverlap;
+		case End:                goto entryEnd;
+	}
+
+entryBegin:
+	/* Set up the next structs at the head of the transition lists. */
+	s1Tel.set( list1 );
+	s2Tel.set( list2 );
+
+	/* Concurrently scan both out ranges. */
+	while ( true ) {
+		if ( s1Tel.trans == 0 ) {
+			/* We are at the end of state1's ranges. Process the rest of
+			 * state2's ranges. */
+			while ( s2Tel.trans != 0 ) {
+				/* Range is only in s2. */
+				CO_RETURN2( ConsumeS2Range, RangeInS2 );
+				s2Tel.increment();
+			}
+			break;
+		}
+		else if ( s2Tel.trans == 0 ) {
+			/* We are at the end of state2's ranges. Process the rest of
+			 * state1's ranges. */
+			while ( s1Tel.trans != 0 ) {
+				/* Range is only in s1. */
+				CO_RETURN2( ConsumeS1Range, RangeInS1 );
+				s1Tel.increment();
+			}
+			break;
+		}
+		/* Both state1's and state2's transition elements are good.
+		 * The signiture of no overlap is a back key being in front of a
+		 * front key. */
+		else if ( s1Tel.highKey < s2Tel.lowKey ) {
+			/* A range exists in state1 that does not overlap with state2. */
+			CO_RETURN2( OnlyInS1Range, RangeInS1 );
+			s1Tel.increment();
+		}
+		else if ( s2Tel.highKey < s1Tel.lowKey ) {
+			/* A range exists in state2 that does not overlap with state1. */
+			CO_RETURN2( OnlyInS2Range, RangeInS2 );
+			s2Tel.increment();
+		}
+		/* There is overlap, must mix the ranges in some way. */
+		else if ( s1Tel.lowKey < s2Tel.lowKey ) {
+			/* Range from state1 sticks out front. Must break it into
+			 * non-overlaping and overlaping segments. */
+			bottomLow = s2Tel.lowKey;
+			bottomHigh = s1Tel.highKey;
+			s1Tel.highKey = s2Tel.lowKey;
+			s1Tel.highKey.decrement();
+			bottomTrans1 = s1Tel.trans;
+
+			/* Notify the caller that we are breaking s1. This gives them a
+			 * chance to duplicate s1Tel[0,1].value. */
+			CO_RETURN2( S1SticksOutBreak, BreakS1 );
+
+			/* Broken off range is only in s1. */
+			CO_RETURN2( S1SticksOut, RangeInS1 );
+
+			/* Advance over the part sticking out front. */
+			s1Tel.lowKey = bottomLow;
+			s1Tel.highKey = bottomHigh;
+			s1Tel.trans = bottomTrans1;
+		}
+		else if ( s2Tel.lowKey < s1Tel.lowKey ) {
+			/* Range from state2 sticks out front. Must break it into
+			 * non-overlaping and overlaping segments. */
+			bottomLow = s1Tel.lowKey;
+			bottomHigh = s2Tel.highKey;
+			s2Tel.highKey = s1Tel.lowKey;
+			s2Tel.highKey.decrement();
+			bottomTrans2 = s2Tel.trans;
+
+			/* Notify the caller that we are breaking s2. This gives them a
+			 * chance to duplicate s2Tel[0,1].value. */
+			CO_RETURN2( S2SticksOutBreak, BreakS2 );
+
+			/* Broken off range is only in s2. */
+			CO_RETURN2( S2SticksOut, RangeInS2 );
+
+			/* Advance over the part sticking out front. */
+			s2Tel.lowKey = bottomLow;
+			s2Tel.highKey = bottomHigh;
+			s2Tel.trans = bottomTrans2;
+		}
+		/* Low ends are even. Are the high ends even? */
+		else if ( s1Tel.highKey < s2Tel.highKey ) {
+			/* Range from state2 goes longer than the range from state1. We
+			 * must break the range from state2 into an evenly overlaping
+			 * segment. */
+			bottomLow = s1Tel.highKey;
+			bottomLow.increment();
+			bottomHigh = s2Tel.highKey;
+			s2Tel.highKey = s1Tel.highKey;
+			bottomTrans2 = s2Tel.trans;
+
+			/* Notify the caller that we are breaking s2. This gives them a
+			 * chance to duplicate s2Tel[0,1].value. */
+			CO_RETURN2( S2DragsBehindBreak, BreakS2 );
+
+			/* Breaking s2 produces exact overlap. */
+			CO_RETURN2( S2DragsBehind, RangeOverlap );
+
+			/* Advance over the front we just broke off of range 2. */
+			s2Tel.lowKey = bottomLow;
+			s2Tel.highKey = bottomHigh;
+			s2Tel.trans = bottomTrans2;
+
+			/* Advance over the entire s1Tel. We have consumed it. */
+			s1Tel.increment();
+		}
+		else if ( s2Tel.highKey < s1Tel.highKey ) {
+			/* Range from state1 goes longer than the range from state2. We
+			 * must break the range from state1 into an evenly overlaping
+			 * segment. */
+			bottomLow = s2Tel.highKey;
+			bottomLow.increment();
+			bottomHigh = s1Tel.highKey;
+			s1Tel.highKey = s2Tel.highKey;
+			bottomTrans1 = s1Tel.trans;
+
+			/* Notify the caller that we are breaking s1. This gives them a
+			 * chance to duplicate s2Tel[0,1].value. */
+			CO_RETURN2( S1DragsBehindBreak, BreakS1 );
+
+			/* Breaking s1 produces exact overlap. */
+			CO_RETURN2( S1DragsBehind, RangeOverlap );
+
+			/* Advance over the front we just broke off of range 1. */
+			s1Tel.lowKey = bottomLow;
+			s1Tel.highKey = bottomHigh;
+			s1Tel.trans = bottomTrans1;
+
+			/* Advance over the entire s2Tel. We have consumed it. */
+			s2Tel.increment();
+		}
+		else {
+			/* There is an exact overlap. */
+			CO_RETURN2( ExactOverlap, RangeOverlap );
+
+			s1Tel.increment();
+			s2Tel.increment();
+		}
+	}
+
+	/* Done, go into end state. */
+	CO_RETURN( End );
+}
+
+
 /* Compare lists of epsilon transitions. Entries are name ids of targets. */
 typedef CmpTable< int, CmpOrd<int> > CmpEpsilonTrans;
 
@@ -779,10 +1005,16 @@ struct FsmAp
 	void longMatchAction( int ordering, LongestMatchPart *lmPart );
 
 	/* Set conditions. */
-	long baseKey( long key, const CondSet &condSet );
-	bool findCondition( long key, Condition *&cond, CondSetMapEl *&csMapEl );
-	CondSetMapEl *addCondition( long key, const CondSet &condSet );
-	void setCondition( Action *cond );
+	CondSpace *addCondSpace( const CondSet &condSet );
+
+	void findEmbedExpansions( ExpansionList &expansionList, 
+		StateAp *destState, Action *condAction );
+	void embedCondition( MergeData &md, StateAp *state, Action *condAction );
+	void embedCondition( StateAp *state, Action *condAction );
+
+	void startFsmCondition( Action *condAction );
+	void allTransCondition( Action *condAction );
+	void leaveFsmCondition( Action *condAction );
 
 	/* Set error actions to execute. */
 	void startErrorAction( int ordering, Action *action, int transferPoint );
@@ -863,7 +1095,7 @@ struct FsmAp
 
 	/* Attach with a new transition. */
 	TransAp *attachNewTrans( StateAp *from, StateAp *to,
-			long onChar1, long onChar2 );
+			Key onChar1, Key onChar2 );
 
 	/* Attach with an existing transition that already in an out list. */
 	void attachTrans( StateAp *from, StateAp *to, TransAp *trans );
@@ -874,9 +1106,6 @@ struct FsmAp
 	/* Detach a transition from a target state. */
 	void detachTrans( StateAp *from, StateAp *to, TransAp *trans );
 
-	/* When a range is split the lower key changes, this fixes it. */
-	void changeRangeLowerKey( TransAp *trans, long oldKey, long newKey );
-
 	/* Detach a state from the graph. */
 	void detachState( StateAp *state );
 
@@ -885,49 +1114,42 @@ struct FsmAp
 	 */
 
 	/* Duplicate a transition that will dropin to a free spot. */
-	TransAp *dupTrans( StateAp *from, TransAp *srcTrans, bool leavingFsm );
-
-	/* In crossing, src trans overwrites the existing one because it has a
-	 * higher priority. */
-	TransAp *overwriteTrans( MergeData &md, StateAp *from,
-			TransAp *destTrans, TransAp *srcTrans, bool leavingFsm );
+	TransAp *dupTrans( StateAp *from, TransAp *srcTrans );
 
 	/* In crossing, two transitions both go to real states. */
 	TransAp *fsmAttachStates( MergeData &md, StateAp *from,
-			TransAp *destTrans, TransAp *srcTrans, bool leavingFsm );
+			TransAp *destTrans, TransAp *srcTrans );
 
 	/* Two transitions are to be crossed, handle the possibility of either
 	 * going to the error state. */
 	TransAp *mergeTrans( MergeData &md, StateAp *from,
-			TransAp *destTrans, TransAp *srcTrans, bool leavingFsm );
+			TransAp *destTrans, TransAp *srcTrans );
 
 	/* Compare deterimne relative priorities of two transition tables. */
 	int comparePrior( const PriorTable &priorTable1, const PriorTable &priorTable2 );
 
 	/* Cross a src transition with one that is already occupying a spot. */
 	TransAp *crossTransitions( MergeData &md, StateAp *from,
-			TransAp *destTrans, TransAp *srcTrans, bool leavingFsm );
+			TransAp *destTrans, TransAp *srcTrans );
 
-	/* Common to copying transition lists and ranges. */
-	TransAp *keyInDestEl( MergeData &md, StateAp *dest, StateAp *src, 
-			NextTrans *destTel, bool leavingFsm );
-	TransAp *keyInSrcEl( MergeData &md, StateAp *dest, StateAp *src, 
-			NextTrans *srcTel, bool leavingFsm );
+	void outTransCopy( MergeData &md, StateAp *dest, TransAp *srcList );
 
-	void appendTrans( TransList &destList, TransAp *trans, bool &covers );
-	void outTransCopy( MergeData &md, StateAp *dest, StateAp *src, bool leavingFsm );
-
-	/* Merging states into a destination state. */
-	void doExpand( StateAp *newState, StateAp *srcState, LongVect *expansion );
-	void removeExpSource( StateAp *destState, LongVect *expansion );
-	void makeCondTrans( StateCondDict &mergedCondDict,
-			Expansions &expansions, StateAp *state );
-	void mergeStates( MergeData &md, StateAp *destState, 
-			StateAp *srcState, bool leavingFsm );
+	void doRemove( MergeData &md, StateAp *destState, ExpansionList &expList1 );
+	void doExpand( MergeData &md, StateAp *destState, ExpansionList &expList1 );
+	void findCondExpInTrans( ExpansionList &expansionList, StateAp *state, 
+			Key lowKey, Key highKey, CondSpace *fromCondSpace, CondSpace *toCondSpace,
+			long destVals, LongVect &toValsList );
+	void findTransExpansions( ExpansionList &expansionList, 
+			StateAp *destState, StateAp *srcState );
+	void findCondExpansions( ExpansionList &expansionList, 
+			StateAp *destState, StateAp *srcState );
+	void mergeStateConds( StateAp *destState, StateAp *srcState );
 
 	/* Merge a set of states into newState. */
 	void mergeStates( MergeData &md, StateAp *destState, 
 			StateAp **srcStates, int numSrc );
+	void mergeStatesLeaving( MergeData &md, StateAp *destState, StateAp *srcState );
+	void mergeStates( MergeData &md, StateAp *destState, StateAp *srcState );
 
 	/* Make all states that are combinations of other states and that
 	 * have not yet had their out transitions filled in. This will 
@@ -964,11 +1186,10 @@ struct FsmAp
 	/* Compare states on data stored in the states. */
 	static int compareStateData( const StateAp *state1, const StateAp *state2 );
 
-	/* This trans is going to be made a leaving-graph-trans from srcState. */
-	void leavingFromState( TransAp *destTrans, StateAp *srcState );
-	
-	/* Invoked when a state looses its final state status. */
-	void relinquishFinal( StateAp *state );
+	/* Out transition data. */
+	void clearOutData( StateAp *state );
+	bool hasOutData( StateAp *state );
+	void transferOutData( StateAp *destState, StateAp *srcState );
 
 	/*
 	 * Allocation.
@@ -981,12 +1202,12 @@ struct FsmAp
 	 * Building basic machines
 	 */
 
-	void concatFsm( long c );
-	void concatFsm( long *str, int len );
-	void concatFsmCI( long *str, int len );
-	void orFsm( long *set, int len );
-	void rangeFsm( long low, long high );
-	void rangeStarFsm( long low, long high );
+	void concatFsm( Key c );
+	void concatFsm( Key *str, int len );
+	void concatFsmCI( Key *str, int len );
+	void orFsm( Key *set, int len );
+	void rangeFsm( Key low, Key high );
+	void rangeStarFsm( Key low, Key high );
 	void emptyFsm( );
 	void lambdaFsm( );
 
@@ -1138,6 +1359,10 @@ struct FsmAp
 	/* Find any states that didn't get marked by the marking algorithm and
 	 * merge them into the primary states of their equivalence class. */
 	void fuseUnmarkedPairs( MarkIndex &markIndex );
+
+	/* Merge neighboring transitions go to the same state and have the same
+	 * transitions data. */
+	void compressTransitions();
 };
 
 
