@@ -37,6 +37,8 @@
 #include "vector.h"
 #include "version.h"
 
+#include "common.cpp"
+
 using std::ostream;
 using std::ofstream;
 using std::cout;
@@ -49,27 +51,36 @@ char *outputFileName = 0;
 
 /* Controls minimization. */
 MinimizeLevel minimizeLevel = MinimizePartition2;
-bool minimizeEveryOp = false;
+MinimizeOpt minimizeOpt = MinimizeMostOps;
 
 /* Graphviz dot file generation. */
 char *machineSpec = 0, *machineName = 0;
 bool machineSpecFound = false;
 
+bool printStatistics = false;
+
 /* Print a summary of the options. */
 void usage()
 {
 	cout <<
 "usage: ragel [options] file\n"
 "general:\n"
-"   -h, -H, -?, --help    Print this usage and exit\n"
-"   -v, --version         Print version information and exit\n"
-"   -o <file>             Write output to <file>\n"
+"   -h, -H, -?, --help   Print this usage and exit\n"
+"   -v, --version        Print version information and exit\n"
+"   -o <file>            Write output to <file>\n"
+"   -s                   Print some statistics on stderr\n"
 "fsm minimization:\n"
-"   -m                    Minimize the number of states (default)\n"
-"   -n                    Do not perform minimization\n"
-"output:\n"
-"   -S <spec>             FSM specification to output for -V\n"
-"   -M <machine>          Machine definition/instantiation to output for -V\n"
+"   -n                   Do not perform minimization\n"
+"   -m                   Minimize at the end of the compilation\n"
+"   -l                   Minimize after most operations (default)\n"
+"   -e                   Minimize after every operation\n"
+"machine selection:\n"
+"   -S <spec>            FSM specification to output for -V\n"
+"   -M <machine>         Machine definition/instantiation to output for -V\n"
+"host language:\n"
+"   -C                   The host language is C, C++, Obj-C or Obj-C++ (default)\n"
+"   -D                   The host language is D\n"
+"   -J                   The host language is Java\n"
 	;	
 }
 
@@ -77,7 +88,7 @@ void usage()
 void version()
 {
 	cout << "Ragel State Machine Compiler version " VERSION << " " PUBDATE << endl <<
-			"Copyright (c) 2001-2005 by Adrian Thurston" << endl;
+			"Copyright (c) 2001-2006 by Adrian Thurston" << endl;
 }
 
 /* Global parse data pointer. */
@@ -99,7 +110,7 @@ ostream &error()
 }
 
 /* Print the opening to an error in the input, then return the error ostream. */
-ostream &error( const BISON_YYLTYPE &loc )
+ostream &error( const YYLTYPE &loc )
 {
 	/* Keep the error count. */
 	if ( id->pd != 0 )
@@ -168,139 +179,10 @@ void escapeLineDirectivePath( std::ostream &out, char *path )
 	}
 }
 
-/* Construct a new parameter checker with for paramSpec. */
-ParamCheck::ParamCheck(char *paramSpec, int argc, char **argv)
-:
-	state(noparam),
-	argOffset(0),
-	curArg(0),
-	iCurArg(1),
-	paramSpec(paramSpec), 
-	argc(argc), 
-	argv(argv)
-{
-}
-
-/* Check a single option. Returns the index of the next parameter.  Sets p to
- * the arg character if valid, 0 otherwise.  Sets parg to the parameter arg if
- * there is one, NULL otherwise. */
-bool ParamCheck::check()
-{
-	bool requiresParam;
-
-	if ( iCurArg >= argc ) {            /* Off the end of the arg list. */
-		state = noparam;
-		return false;
-	}
-
-	if ( argOffset != 0 && *argOffset == 0 ) {
-		/* We are at the end of an arg string. */
-		iCurArg += 1;
-		if ( iCurArg >= argc ) {
-			state = noparam;
-			return false;
-		}
-		argOffset = 0;
-	}
-
-	if ( argOffset == 0 ) {
-		/* Set the current arg. */
-		curArg = argv[iCurArg];
-
-		/* We are at the beginning of an arg string. */
-		if ( argv[iCurArg] == 0 ||        /* Argv[iCurArg] is null. */
-			 argv[iCurArg][0] != '-' ||   /* Not a param. */
-			 argv[iCurArg][1] == 0 ) {    /* Only a dash. */
-			parameter = 0;
-			parameterArg = 0;
-
-			iCurArg += 1;
-			state = noparam;
-			return true;
-		}
-		argOffset = argv[iCurArg] + 1;
-	}
-
-	/* Get the arg char. */
-	char argChar = *argOffset;
-	
-	/* Loop over all the parms and look for a match. */
-	char *pSpec = paramSpec;
-	while ( *pSpec != 0 ) {
-		char pSpecChar = *pSpec;
-
-		/* If there is a ':' following the char then
-		 * it requires a parm.  If a parm is required
-		 * then move ahead two in the parmspec. Otherwise
-		 * move ahead one in the parm spec. */
-		if ( pSpec[1] == ':' ) {
-			requiresParam = true;
-			pSpec += 2;
-		}
-		else {
-			requiresParam = false;
-			pSpec += 1;
-		}
-
-		/* Do we have a match. */
-		if ( argChar == pSpecChar ) {
-			if ( requiresParam ) {
-				if ( argOffset[1] == 0 ) {
-					/* The param must follow. */
-					if ( iCurArg + 1 == argc ) {
-						/* We are the last arg so there
-						 * cannot be a parameter to it. */
-						parameter = argChar;
-						parameterArg = 0;
-						iCurArg += 1;
-						argOffset = 0;
-						state = invalid;
-						return true;
-					}
-					else {
-						/* the parameter to the arg is the next arg. */
-						parameter = pSpecChar;
-						parameterArg = argv[iCurArg + 1];
-						iCurArg += 2;
-						argOffset = 0;
-						state = match;
-						return true;
-					}
-				}
-				else {
-					/* The param for the arg is built in. */
-					parameter = pSpecChar;
-					parameterArg = argOffset + 1;
-					iCurArg += 1;
-					argOffset = 0;
-					state = match;
-					return true;
-				}
-			}
-			else {
-				/* Good, we matched the parm and no
-				 * arg is required. */
-				parameter = pSpecChar;
-				parameterArg = 0;
-				argOffset += 1;
-				state = match;
-				return true;
-			}
-		}
-	}
-
-	/* We did not find a match. Bad Argument. */
-	parameter = argChar;
-	parameterArg = 0;
-	argOffset += 1;
-	state = invalid;
-	return true;
-}
-
 /* Main, process args and call yyparse to start scanning input. */
 int main(int argc, char **argv)
 {
-	ParamCheck pc("o:enmabjkS:M:vHh?-:", argc, argv);
+	ParamCheck pc("o:nmleabjkS:M:CDJvHh?-:s", argc, argv);
 	char *inputFileName = 0;
 
 	while ( pc.check() ) {
@@ -309,16 +191,10 @@ int main(int argc, char **argv)
 			switch ( pc.parameter ) {
 			/* Output. */
 			case 'o':
-				if ( *pc.parameterArg == 0 ) {
-					/* Complain, someone used -o "" */
-					error() << "zero length output file name" << endl;
-					exit(1);
-				}
-				else if ( outputFileName != 0 ) {
-					/* Complain, two output files given. */
-					error() << "more than one output file" << endl;
-					exit(1);
-				}
+				if ( *pc.parameterArg == 0 )
+					error() << "a zero length output file name was given" << endl;
+				else if ( outputFileName != 0 )
+					error() << "more than one output file name was given" << endl;
 				else {
 					/* Ok, remember the output file name. */
 					outputFileName = pc.parameterArg;
@@ -326,14 +202,17 @@ int main(int argc, char **argv)
 				break;
 
 			/* Minimization, mostly hidden options. */
-			case 'e':
-				minimizeEveryOp = true;
-				break;
 			case 'n':
-				minimizeLevel = MinimizeNone;
+				minimizeOpt = MinimizeNone;
 				break;
 			case 'm':
-				minimizeLevel = MinimizePartition2;
+				minimizeOpt = MinimizeEnd;
+				break;
+			case 'l':
+				minimizeOpt = MinimizeMostOps;
+				break;
+			case 'e':
+				minimizeOpt = MinimizeEveryOp;
 				break;
 			case 'a':
 				minimizeLevel = MinimizeApprox;
@@ -350,15 +229,10 @@ int main(int argc, char **argv)
 
 			/* Machine spec. */
 			case 'S':
-				if ( *pc.parameterArg == 0 ) {
-					/* Complain, need a graph spec. */
+				if ( *pc.parameterArg == 0 )
 					error() << "please specify an argument to -S" << endl;
-					exit(1);
-				}
-				else if ( machineSpec != 0 ) {
-					error() << "FSM specification to generate already given" << endl;
-					exit(1);
-				}
+				else if ( machineSpec != 0 )
+					error() << "more than one -S argument was given" << endl;
 				else {
 					/* Ok, remember the path to the machine to generate. */
 					machineSpec = pc.parameterArg;
@@ -367,21 +241,30 @@ int main(int argc, char **argv)
 
 			/* Machine path. */
 			case 'M':
-				if ( *pc.parameterArg == 0 ) {
-					/* Complain, need a graph spec. */
+				if ( *pc.parameterArg == 0 )
 					error() << "please specify an argument to -M" << endl;
-					exit(1);
-				}
-				else if ( machineName != 0 ) {
-					error() << "machine to generate already given" << endl;
-					exit(1);
-				}
+				else if ( machineName != 0 )
+					error() << "more than one -M argument was given" << endl;
 				else {
 					/* Ok, remember the machine name to generate. */
 					machineName = pc.parameterArg;
 				}
 				break;
 
+			/* Host language types. */
+			case 'C':
+				hostLangType = CCode;
+				hostLang = &hostLangC;
+				break;
+			case 'D':
+				hostLangType = DCode;
+				hostLang = &hostLangD;
+				break;
+			case 'J':
+				hostLangType = JavaCode;
+				hostLang = &hostLangJava;
+				break;
+
 			/* Version and help. */
 			case 'v':
 				version();
@@ -389,6 +272,9 @@ int main(int argc, char **argv)
 			case 'H': case 'h': case '?':
 				usage();
 				exit(0);
+			case 's':
+				printStatistics = true;
+				break;
 			case '-':
 				if ( strcasecmp(pc.parameterArg, "help") == 0 ) {
 					usage();
@@ -399,38 +285,30 @@ int main(int argc, char **argv)
 					exit(0);
 				}
 				else {
-					error() << "invalid parameter" << endl;
-					exit(1);
+					error() << "--" << pc.parameterArg << 
+							" is an invalid argument" << endl;
 				}
 			}
 			break;
 
 		case ParamCheck::invalid:
-			error() << "invalid parameter" << endl;
-			exit(1);
+			error() << "-" << pc.parameter << " is an invalid argument" << endl;
+			break;
 
 		case ParamCheck::noparam:
 			/* It is interpreted as an input file. */
-			if ( *pc.curArg == 0 ) {
-				error() << "zero length input file name" << endl;
-				exit(1);
+			if ( *pc.curArg == 0 )
+				error() << "a zero length input file name was given" << endl;
+			else if ( inputFileName != 0 )
+				error() << "more than one input file name was given" << endl;
+			else {
+				/* OK, Remember the filename. */
+				inputFileName = pc.curArg;
 			}
-			if ( inputFileName != 0 ) {
-				error() << "more than one input file" << endl;
-				exit(1);
-			}
-			/* Remember the filename. */
-			inputFileName = pc.curArg;
 			break;
 		}
 	}
 
-	/* If the user wants minimization at every op, but no minimization level 
-	 * is given, then print a warning. */
-	if ( minimizeEveryOp && minimizeLevel == MinimizeNone )
-		warning() << "-e given but minimization is not enabled" << endl;
-
-
 	/* Bail on above errors. */
 	if ( gblErrorCount > 0 )
 		exit(1);
