@@ -23,6 +23,7 @@
 #include <iomanip>
 #include <errno.h>
 #include <stdlib.h>
+#include <limits.h>
 
 #include "ragel.h"
 #include "rlparse.h"
@@ -35,7 +36,7 @@ using namespace std;
 
 char machineMain[] = "main";
 
-InputLoc::InputLoc( const BISON_YYLTYPE &loc )
+InputLoc::InputLoc( const YYLTYPE &loc )
 :
 	line(loc.first_line), 
 	col(loc.first_column)
@@ -63,21 +64,17 @@ void TokenData::init( char c )
 
 /* Perform minimization after an operation according 
  * to the command line args. */
-void afterOpMinimize( FsmAp *fsm )
+void afterOpMinimize( FsmAp *fsm, bool lastInSeq )
 {
 	/* Switch on the prefered minimization algorithm. */
-	if ( minimizeEveryOp ) {
-		if ( minimizeLevel != MinimizeNone ) {
-			/* First clean up the graph. FsmAp operations may leave these
-			 * lying around. There should be no dead end states. The subtract
-			 * intersection operators are the only places where they may be
-			 * created and those operators clean them up. */
-			fsm->removeUnreachableStates();
-		}
+	if ( minimizeOpt == MinimizeEveryOp || minimizeOpt == MinimizeMostOps && lastInSeq ) {
+		/* First clean up the graph. FsmAp operations may leave these
+		 * lying around. There should be no dead end states. The subtract
+		 * intersection operators are the only places where they may be
+		 * created and those operators clean them up. */
+		fsm->removeUnreachableStates();
 
 		switch ( minimizeLevel ) {
-			case MinimizeNone:
-				break;
 			case MinimizeApprox:
 				fsm->minimizeApproximate();
 				break;
@@ -106,292 +103,59 @@ int countTransitions( FsmAp *fsm )
 	return numTrans;
 }
 
-long makeFsmKeyHex( char *str, const InputLoc &loc, ParseData *pd )
+Key makeFsmKeyHex( char *str, const InputLoc &loc, ParseData *pd )
 {
-	/* Convert the number to a decimal. Reset errno so we can check for
-	 * overflow or underflow. In the event of an error, sets the return val to
-	 * the upper or lower bound being tested against. */
+	/* Reset errno so we can check for overflow or underflow. In the event of
+	 * an error, sets the return val to the upper or lower bound being tested
+	 * against. */
 	errno = 0;
+	unsigned int size = keyOps->alphType->size;
+	bool unusedBits = size < sizeof(unsigned long);
+
 	unsigned long ul = strtoul( str, 0, 16 );
 
-	switch ( pd->alphType ) {
-	case AT_Char: {
-		/* Check against upper bound. */
-		if ( ul > RL_UCHAR_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows char alphabet" << endl;
-			ul = RL_UCHAR_MAX;
-		}
-
-		/* Cast to a char to get the proper sign, then extend to an integer. */
-		return (char)ul;
+	if ( errno == ERANGE || unusedBits && ul >> (size * 8) ) {
+		error(loc) << "literal " << str << " overflows the alphabet type" << endl;
+		ul = 1 << (size * 8);
 	}
 
-	case AT_UnsignedChar: {
-		/* Check against upper bound. */
-		if ( ul > RL_UCHAR_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows unsigned char alphabet" << endl;
-			ul = RL_UCHAR_MAX;
-		}
+	if ( unusedBits && keyOps->alphType->isSigned && ul >> (size * 8 - 1) )
+		ul |= (0xffffffff >> (size*8 ) ) << (size*8);
 
-		/* Store cast to unsigned char, then extend to an integer. */
-		return (unsigned char)ul;
-	}
-
-	case AT_Short: {
-		/* Check against upper bound. */
-		if ( ul > RL_USHORT_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows short alphabet" << endl;
-			ul = RL_USHORT_MAX;
-		}
-
-		/* Cast to a short to get the proper sign, then extend to an integer. */
-		return (short)ul;
-	}
-
-	case AT_UnsignedShort: {
-		/* Check against upper bound. */
-		if ( ul > RL_USHORT_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows unsigned short alphabet" << endl;
-			ul = RL_USHORT_MAX;
-		}
-
-		/* Store cast to unsigned short, then extend to an integer. */
-		return (unsigned short)ul;
-	}
-
-	case AT_Int: {
-		/* Check against upper bound. */
-		if ( ul > RL_UINT_MAX || errno == ERANGE && ul == ULONG_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows int alphabet" << endl;
-			ul = ULONG_MAX;
-		}
-
-		/* Store in integer format. */
-		return (int)ul;
-	}
-
-	case AT_UnsignedInt: {
-		/* Check against upper bound. */
-		if ( ul > RL_UINT_MAX || errno == ERANGE && ul == ULONG_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows unsigned int alphabet" << endl;
-			ul = ULONG_MAX;
-		}
-
-		/* Store in integer format. */
-		return (unsigned int)ul;
-	}
-
-	case AT_Long: {
-		/* Check against upper bound. */
-		if ( ul > RL_ULONG_MAX || errno == ERANGE && ul == ULONG_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows long alphabet" << endl;
-			ul = ULONG_MAX;
-		}
-
-		return (long)ul;
-	}
-
-	case AT_UnsignedLong: {
-		/* Check against upper bound. */
-		if ( ul > RL_ULONG_MAX || errno == ERANGE && ul == ULONG_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows unsigned long alphabet" << endl;
-			ul = ULONG_MAX;
-		}
-
-		return (unsigned long)ul;
-	}}
-	/* Not reached. */
-	return 0;
+	return Key( (long)ul );
 }
 
-long makeFsmKeyDec( char *str, const InputLoc &loc, ParseData *pd )
+Key makeFsmKeyDec( char *str, const InputLoc &loc, ParseData *pd )
 {
-	switch ( pd->alphType ) {
-	case AT_Char: {
-		/* Convert the number to a decimal. If there is overflow or underflow
-		 * the num will be set to LONG_MIN/LONG_MAX which are way too low and
-		 * high for char anyways so no need to check. */
-		long l = strtol( str, 0, 10 );
+	/* Convert the number to a decimal. First reset errno so we can check
+	 * for overflow or underflow. */
+	errno = 0;
+	long long minVal = keyOps->alphType->minVal;
+	long long maxVal = keyOps->alphType->maxVal;
 
-		/* Check against lower bound. */
-		if ( l < RL_CHAR_MIN ) {
-	 		error(loc) << "literal " << str << 
-					" underflows char alphabet" << endl;
-			l = RL_CHAR_MIN;
-		}
+	long long ll = strtoll( str, 0, 10 );
 
-		/* Check against upper bound. */
-		if ( l > RL_CHAR_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows char alphabet" << endl;
-			l = RL_CHAR_MAX;
-		}
-		return l;
+	/* Check for underflow. */
+	if ( errno == ERANGE && ll < 0 || ll < minVal) {
+		error(loc) << "literal " << str << " underflows the alphabet type" << endl;
+		ll = minVal;
+	}
+	/* Check for overflow. */
+	else if ( errno == ERANGE && ll > 0 || ll > maxVal ) {
+		error(loc) << "literal " << str << " overflows the alphabet type" << endl;
+		ll = maxVal;
 	}
 
-	case AT_UnsignedChar: {
-		/* Can't allow a negative decimal number for unsigned char type. */
-		if ( str[0] == '-' ) {
-			/* Recover by advancing over the dash. */
-			error(loc) << "literal " << str << 
-					" underflows unsigned char alphabet" << endl;
-			str += 1;
-		}
-
-		/* Convert to an unsigned number and bounds check it. */
-		unsigned long ul = strtoul( str, 0, 10 );
-		if ( ul > RL_UCHAR_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows unsigned char alphabet" << endl;
-			ul = RL_UCHAR_MAX;
-		}
-		return ul;
-	}
-
-	case AT_Short: {
-		/* Convert the number to a decimal. If there is overflow or underflow
-		 * the num will be set to the high or low limits which are exceed the
-		 * bounds for short so no need to check. */
-		long l = strtol( str, 0, 10 );
-
-		/* Check against lower bound. */
-		if ( l < RL_SHORT_MIN ) {
-	 		error(loc) << "literal " << str << 
-					" underflows short alphabet" << endl;
-			l = RL_SHORT_MIN;
-		}
-
-		/* Check against upper bound. */
-		if ( l > RL_SHORT_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows short alphabet" << endl;
-			l = RL_SHORT_MAX;
-		}
-		return l;
-	}
-
-	case AT_UnsignedShort: {
-		/* Can't allow a negative decimal number for unsigned short type. */
-		if ( str[0] == '-' ) {
-			/* Recover by skippping the dash. */
-			error(loc) << "literal " << str << 
-					" underflows unsigned short alphabet" << endl;
-			str += 1;
-		}
-
-		/* Convert to an unsigned number and bounds check it. */
-		unsigned long ul = strtoul( str, 0, 10 );
-		if ( ul > RL_USHORT_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows unsigned short alphabet" << endl;
-			ul = RL_USHORT_MAX;
-		}
-		return ul;
-	}
-
-	case AT_Int: {
-		/* Convert the number to a decimal. First reset errno so we can check
-		 * for overflow or underflow. */
-		errno = 0;
-		long l = strtol( str, 0, 10 );
-
-		/* Check lower bound. */
-		if ( l < RL_INT_MIN || errno == ERANGE && l == LONG_MIN ) {
-	 		error(loc) << "literal " << str << 
-					" underflows int alphabet" << endl;
-			l = LONG_MIN;
-		}
-
-		/* Check upper bound. */
-		if ( l > RL_INT_MAX || errno == ERANGE && l == LONG_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows int alphabet" << endl;
-			l = LONG_MAX;
-		}
-		return l;
-	}
-
-	case AT_UnsignedInt: {
-		/* Cannot allow negative numbers for an unsigned type. */
-		if ( str[0] == '-' ) {
-			/* Recover by skipping the dash. */
-			error(loc) << "literal " << str << 
-					" underflows unsigned int alphabet" << endl;
-			str += 1;
-		}
-
-		/* Convert to an unsigned number and bounds check it. */
-		errno = 0;
-		unsigned long ul = strtoul( str, 0, 10 );
-		if ( ul > RL_UINT_MAX || errno == ERANGE && ul == ULONG_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows unsigned int alphabet" << endl;
-			ul = ULONG_MAX;
-		}
-
-		/* Store in integer format. */
-		return ul;
-	}
-
-	case AT_Long: {
-		/* Convert the number to a decimal. First reset errno so we can check
-		 * for overflow or underflow. */
-		errno = 0;
-		long l = strtol( str, 0, 10 );
-
-		/* Check lower bound. */
-		if ( l < RL_LONG_MIN || errno == ERANGE && l == LONG_MIN ) {
-	 		error(loc) << "literal " << str << 
-					" underflows long alphabet" << endl;
-			l = LONG_MIN;
-		}
-
-		/* Check upper bound. */
-		if ( l > RL_LONG_MAX || errno == ERANGE && l == LONG_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows long alphabet" << endl;
-			l = LONG_MAX;
-		}
-		return l;
-	}
-
-	case AT_UnsignedLong: {
-		/* Cannot allow negative numbers for an unsigned type. */
-		if ( str[0] == '-' ) {
-			/* Recover by skipping the dash. */
-			error(loc) << "literal " << str << 
-					" underflows unsigned long alphabet" << endl;
-			str += 1;
-		}
-
-		/* Convert to an unsigned number and bounds check it. */
-		errno = 0;
-		unsigned long ul = strtoul( str, 0, 10 );
-		if ( ul > RL_ULONG_MAX || errno == ERANGE && ul == ULONG_MAX ) {
-			error(loc) << "literal " << str << 
-					" overflows unsigned long alphabet" << endl;
-			ul = ULONG_MAX;
-		}
-
-		/* Store in integer format. */
-		return ul;
-	}}
-	/* Not reached. */
-	return 0;
+	if ( keyOps->alphType->isSigned )
+		return Key( (long)ll );
+	else
+		return Key( (unsigned long)ll );
 }
 
 /* Make an fsm key in int format (what the fsm graph uses) from an alphabet
  * number returned by the parser. Validates that the number doesn't overflow
  * the alphabet type. */
-long makeFsmKeyNum( char *str, const InputLoc &loc, ParseData *pd )
+Key makeFsmKeyNum( char *str, const InputLoc &loc, ParseData *pd )
 {
 	/* Switch on hex/decimal format. */
 	if ( str[0] == '0' && str[1] == 'x' )
@@ -403,56 +167,54 @@ long makeFsmKeyNum( char *str, const InputLoc &loc, ParseData *pd )
 /* Make an fsm int format (what the fsm graph uses) from a single character.
  * Performs proper conversion depending on signed/unsigned property of the
  * alphabet. */
-long makeFsmKeyChar( char c, ParseData *pd )
+Key makeFsmKeyChar( char c, ParseData *pd )
 {
-	long retVal;
-	if ( pd->isAlphSigned() ) {
+	if ( keyOps->isSigned ) {
 		/* Copy from a char type. */
-		retVal = c;
+		return Key( c );
 	}
 	else {
 		/* Copy from an unsigned byte type. */
-		retVal = (unsigned char)c;
+		return Key( (unsigned char)c );
 	}
-	return retVal;
 }
 
 /* Make an fsm key array in int format (what the fsm graph uses) from a string
  * of characters. Performs proper conversion depending on signed/unsigned
  * property of the alphabet. */
-void makeFsmKeyArray( long *result, char *data, int len, ParseData *pd )
+void makeFsmKeyArray( Key *result, char *data, int len, ParseData *pd )
 {
-	if ( pd->isAlphSigned() ) {
+	if ( keyOps->isSigned ) {
 		/* Copy from a char star type. */
 		char *src = data;
 		for ( int i = 0; i < len; i++ )
-			result[i] = src[i];
+			result[i] = Key(src[i]);
 	}
 	else {
 		/* Copy from an unsigned byte ptr type. */
 		unsigned char *src = (unsigned char*) data;
 		for ( int i = 0; i < len; i++ )
-			result[i] = src[i];
+			result[i] = Key(src[i]);
 	}
 }
 
 /* Like makeFsmKeyArray except the result has only unique keys. They ordering
  * will be changed. */
-long *makeFsmUniqueKeyArray( int &resLength, char *data, int len, 
+void makeFsmUniqueKeyArray( KeySet &result, char *data, int len, 
 		bool caseInsensitive, ParseData *pd )
 {
 	/* Use a transitions list for getting unique keys. */
-	KeySet keySet;
-	if ( pd->isAlphSigned() ) {
+	if ( keyOps->isSigned ) {
 		/* Copy from a char star type. */
 		char *src = data;
 		for ( int si = 0; si < len; si++ ) {
-			keySet.insert( src[si] );
+			Key key( src[si] );
+			result.insert( key );
 			if ( caseInsensitive ) {
-				if ( 'a' <= src[si] && src[si] <= 'z' )
-					keySet.insert( 'A' + ( src[si] - 'a' ) );
-				else if ( 'A' <= src[si] && src[si] <= 'Z' )
-					keySet.insert( 'a' + ( src[si] - 'A' ) );
+				if ( key.isLower() )
+					result.insert( key.toUpper() );
+				else if ( key.isUpper() )
+					result.insert( key.toLower() );
 			}
 		}
 	}
@@ -460,84 +222,29 @@ long *makeFsmUniqueKeyArray( int &resLength, char *data, int len,
 		/* Copy from an unsigned byte ptr type. */
 		unsigned char *src = (unsigned char*) data;
 		for ( int si = 0; si < len; si++ ) {
-			keySet.insert( src[si] );
+			Key key( src[si] );
+			result.insert( key );
 			if ( caseInsensitive ) {
-				if ( 'a' <= src[si] && src[si] <= 'z' )
-					keySet.insert( 'A' + ( src[si] - 'a' ) );
-				else if ( 'A' <= src[si] && src[si] <= 'Z' )
-					keySet.insert( 'a' + ( src[si] - 'A' ) );
+				if ( key.isLower() )
+					result.insert( key.toUpper() );
+				else if ( key.isUpper() )
+					result.insert( key.toLower() );
 			}
 		}
 	}
-
-	/* Copy the data back in and return the number of unique keys written. */
-	resLength = keySet.length();
-	long *result = new long[keySet.length()];
-	memcpy( result, keySet.data, sizeof(long)*keySet.length() );
-	return result;
 }
 
 FsmAp *dotFsm( ParseData *pd )
 {
 	FsmAp *retFsm = new FsmAp();
-	switch ( pd->alphType ) {
-		case AT_Char:
-			retFsm->rangeFsm( RL_CHAR_MIN, RL_CHAR_MAX );
-			break;
-		case AT_UnsignedChar:
-			retFsm->rangeFsm( 0, RL_UCHAR_MAX );
-			break;
-		case AT_Short:
-			retFsm->rangeFsm( RL_SHORT_MIN, RL_SHORT_MAX );
-			break;
-		case AT_UnsignedShort:
-			retFsm->rangeFsm( 0, RL_USHORT_MAX );
-			break;
-		case AT_Int:
-			retFsm->rangeFsm( RL_INT_MIN, RL_INT_MAX );
-			break;
-		case AT_UnsignedInt:
-			retFsm->rangeFsm( 0, RL_UINT_MAX );
-			break;
-		case AT_Long:
-			retFsm->rangeFsm( RL_LONG_MIN, RL_LONG_MAX );
-			break;
-		case AT_UnsignedLong:
-			retFsm->rangeFsm( 0, RL_ULONG_MAX );
-			break;
-	}
+	retFsm->rangeFsm( keyOps->minKey, keyOps->maxKey );
 	return retFsm;
 }
 
 FsmAp *dotStarFsm( ParseData *pd )
 {
 	FsmAp *retFsm = new FsmAp();
-	switch ( pd->alphType ) {
-		case AT_Char:
-			retFsm->rangeStarFsm( RL_CHAR_MIN, RL_CHAR_MAX );
-			break;
-		case AT_UnsignedChar:
-			retFsm->rangeStarFsm( 0, RL_UCHAR_MAX );
-			break;
-		case AT_Short:
-			retFsm->rangeStarFsm( RL_SHORT_MIN, RL_SHORT_MAX );
-			break;
-		case AT_UnsignedShort:
-			retFsm->rangeStarFsm( 0, RL_USHORT_MAX );
-			break;
-		case AT_Int:
-			retFsm->rangeStarFsm( RL_INT_MIN, RL_INT_MAX );
-			break;
-		case AT_UnsignedInt:
-			retFsm->rangeStarFsm( 0, RL_UINT_MAX );
-			break;
-		case AT_Long:
-			retFsm->rangeStarFsm( RL_LONG_MIN, RL_LONG_MAX );
-			break;
-		case AT_UnsignedLong:
-			retFsm->rangeStarFsm( 0, RL_ULONG_MAX );
-			break;
-	}
+	retFsm->rangeStarFsm( keyOps->minKey, keyOps->maxKey );
 	return retFsm;
 }
 
@@ -546,7 +253,7 @@ FsmAp *makeBuiltin( BuiltinMachine builtin, ParseData *pd )
 {
 	/* FsmAp created to return. */
 	FsmAp *retFsm = 0;
-	bool isSigned = pd->isAlphSigned();
+	bool isSigned = keyOps->isSigned;
 
 	switch ( builtin ) {
 	case BT_Any: {
@@ -564,9 +271,7 @@ FsmAp *makeBuiltin( BuiltinMachine builtin, ParseData *pd )
 		/* Ascii extended characters. This is the full byte range. Dependent
 		 * on signed, vs no signed. If the alphabet is one byte then just use
 		 * dot fsm. */
-		if ( pd->alphType == AT_Char || pd->alphType == AT_UnsignedChar )
-			retFsm = dotFsm( pd );
-		else if ( isSigned ) {
+		if ( isSigned ) {
 			retFsm = new FsmAp();
 			retFsm->rangeFsm( -128, 127 );
 		}
@@ -726,7 +431,6 @@ ParseData::ParseData( char *fileName, char *fsmName,
 	/* 0 is reserved for global error actions. */
 	nextLocalErrKey(1),
 	nextNameId(0),
-	alphType(AT_Char),
 	alphTypeSet(false),
 	getKeyExpr(0),
 	accessExpr(0),
@@ -1100,13 +804,6 @@ void ParseData::makeNameTree( GraphDictEl *dictEl )
 	fillNameIndex( rootName );
 }
 
-
-/* Is the alphabet type a signed type? */
-bool ParseData::isAlphSigned()
-{
-	return alphType == AT_Char || alphType == AT_Short || alphType == AT_Int;
-}
-
 void ParseData::createBuiltin( char *name, BuiltinMachine builtin )
 {
 	Expression *expression = new Expression( builtin );
@@ -1143,24 +840,17 @@ void ParseData::initGraphDict( )
 bool ParseData::setAlphType( char *s1, char *s2 )
 {
 	bool valid = false;
-	if ( strcmp( s1, "unsigned" ) == 0 ) {
-		if ( strcmp( s2, "char" ) == 0 ) {
-			alphType = AT_UnsignedChar;
-			valid = true;
-		}
-		else if ( strcmp( s2, "short" ) == 0 ) {
-			alphType = AT_UnsignedShort;
-			valid = true;
-		}
-		else if ( strcmp( s2, "int" ) == 0 ) {
-			alphType = AT_UnsignedInt;
-			valid = true;
-		}
-		else if ( strcmp( s2, "long" ) == 0 ) {
-			alphType = AT_UnsignedLong;
+	for ( int i = 0; i < hostLang->numHostTypes; i++ ) {
+		if ( strcmp( s1, hostLang->hostTypes[i].data1 ) == 0 && 
+				hostLang->hostTypes[i].data2 != 0 && 
+				strcmp( s2, hostLang->hostTypes[i].data2 ) == 0 )
+		{
 			valid = true;
+			userAlphType = hostLang->hostTypes + i;
+			break;
 		}
 	}
+
 	alphTypeSet = true;
 	return valid;
 }
@@ -1169,70 +859,18 @@ bool ParseData::setAlphType( char *s1, char *s2 )
 bool ParseData::setAlphType( char *s1 )
 {
 	bool valid = false;
-	if ( strcmp( s1, "char" ) == 0 ) {
-		alphType = AT_Char;
-		valid = true;
-	}
-	else if ( strcmp( s1, "short" ) == 0 ) {
-		alphType = AT_Short;
-		valid = true;
-	}
-	else if ( strcmp( s1, "int" ) == 0 ) {
-		alphType = AT_Int;
-		valid = true;
-	}
-	else if ( strcmp( s1, "long" ) == 0 ) {
-		alphType = AT_Long;
-		valid = true;
-	}
-	alphTypeSet = true;
-	return valid;
-}
-
-void ParseData::setLowerUpperRange( )
-{
-	if ( lowerNum != 0 ) {
-		/* If ranges are given then interpret the alphabet type. */
-		lowKey = makeFsmKeyNum( lowerNum, rangeLowLoc, this );
-		highKey = makeFsmKeyNum( upperNum, rangeHighLoc, this );
-	}
-	else {
-		/* Use default ranges for alphabet type. */
-		switch ( alphType ) {
-		case AT_Char:
-			lowKey = RL_CHAR_MIN;
-			highKey = RL_CHAR_MAX;
-			break;
-		case AT_UnsignedChar:
-			lowKey = RL_UCHAR_MIN;
-			highKey = RL_UCHAR_MAX;
-			break;
-		case AT_Short:
-			lowKey = RL_SHORT_MIN;
-			highKey = RL_SHORT_MAX;
-			break;
-		case AT_UnsignedShort:
-			lowKey = RL_USHORT_MIN;
-			highKey = RL_USHORT_MAX;
-			break;
-		case AT_Int:
-			lowKey = RL_INT_MIN;
-			highKey = RL_INT_MAX;
-			break;
-		case AT_UnsignedInt:
-			lowKey = RL_UINT_MIN;
-			highKey = RL_UINT_MAX;
-			break;
-		case AT_Long:
-			lowKey = RL_LONG_MIN;
-			highKey = RL_LONG_MAX;
-			break;
-		case AT_UnsignedLong:
-			lowKey = RL_ULONG_MIN;
-			highKey = RL_ULONG_MAX;
+	for ( int i = 0; i < hostLang->numHostTypes; i++ ) {
+		if ( strcmp( s1, hostLang->hostTypes[i].data1 ) == 0 && 
+				hostLang->hostTypes[i].data2 == 0 )
+		{
+			valid = true;
+			userAlphType = hostLang->hostTypes + i;
 			break;
 		}
 	}
+
+	alphTypeSet = true;
+	return valid;
 }
 
 /* Initialize the key operators object that will be referenced by all fsms
@@ -1240,12 +878,17 @@ void ParseData::setLowerUpperRange( )
 void ParseData::initKeyOps( )
 {
 	/* Signedness and bounds. */
-	thisKeyOps.isAlphSigned = isAlphSigned();
-	thisKeyOps.lowKey = lowKey;
-	thisKeyOps.highKey = highKey;
+	HostType *alphType = alphTypeSet ? userAlphType : hostLang->defaultAlphType;
+	thisKeyOps.setAlphType( alphType );
 
-	thisCondData.nextCondKey = highKey;
-	keyOps->inc( thisCondData.nextCondKey );
+	if ( lowerNum != 0 ) {
+		/* If ranges are given then interpret the alphabet type. */
+		thisKeyOps.minKey = makeFsmKeyNum( lowerNum, rangeLowLoc, this );
+		thisKeyOps.maxKey = makeFsmKeyNum( upperNum, rangeHighLoc, this );
+	}
+
+	thisCondData.nextCondKey = thisKeyOps.maxKey;
+	thisCondData.nextCondKey.increment();
 }
 
 void ParseData::printNameInst( NameInst *nameInst, int level )
@@ -1409,25 +1052,27 @@ FsmAp *ParseData::makeInstance( GraphDictEl *gdNode )
 	 * because they will just hinder minimization as well. Clear them. */
 	graph->clearAllPriorities();
 
-	/* Minimize here even if we minimized at every op. Now that function
-	 * keys have been cleared we may get a more minimal fsm. */
-	switch ( minimizeLevel ) {
-		case MinimizeNone:
-			break;
-		case MinimizeApprox:
-			graph->minimizeApproximate();
-			break;
-		case MinimizeStable:
-			graph->minimizeStable();
-			break;
-		case MinimizePartition1:
-			graph->minimizePartition1();
-			break;
-		case MinimizePartition2:
-			graph->minimizePartition2();
-			break;
+	if ( minimizeOpt != MinimizeNone ) {
+		/* Minimize here even if we minimized at every op. Now that function
+		 * keys have been cleared we may get a more minimal fsm. */
+		switch ( minimizeLevel ) {
+			case MinimizeApprox:
+				graph->minimizeApproximate();
+				break;
+			case MinimizeStable:
+				graph->minimizeStable();
+				break;
+			case MinimizePartition1:
+				graph->minimizePartition1();
+				break;
+			case MinimizePartition2:
+				graph->minimizePartition2();
+				break;
+		}
 	}
 
+	graph->compressTransitions();
+
 	return graph;
 }
 
@@ -1519,6 +1164,7 @@ FsmAp *ParseData::makeAll()
 
 void ParseData::analyzeAction( Action *action, InlineList *inlineList )
 {
+	/* FIXME: Actions used as conditions should be very constrained. */
 	for ( InlineList::Iter item = *inlineList; item.lte(); item++ ) {
 		if ( item->type == InlineItem::Call || item->type == InlineItem::CallExpr )
 			action->anyCall = true;
@@ -1592,22 +1238,18 @@ void ParseData::checkInlineList( Action *act, InlineList *inlineList )
 
 void ParseData::checkAction( Action *action )
 {
-	if ( action->numTransRefs > 0 || action->numToStateRefs > 0 ||
-			action->numFromStateRefs > 0 || action->numEofRefs > 0 )
-	{
-		if ( action->anyCall && !action->isLmAction ) {
-			/* Check for actions with calls that are embedded within a longest match
-			 * machine. */
-			for ( ActionRefs::Iter ar = action->actionRefs; ar.lte(); ar++ ) {
-				NameInst *check = *ar;
-				while ( check != 0 ) {
-					if ( check->isLongestMatch ) {
-						error(action->loc) << "fcall permitted only in pattern actions "
-							"of a longest match" << endl;
-						break;
-					}
-					check = check->parent;
+	if ( !action->isLmAction && action->numRefs() > 0 && action->anyCall ) {
+		/* Check for actions with calls that are embedded within a longest match
+		 * machine. */
+		for ( ActionRefs::Iter ar = action->actionRefs; ar.lte(); ar++ ) {
+			NameInst *check = *ar;
+			while ( check != 0 ) {
+				if ( check->isLongestMatch ) {
+					error(action->loc) << "fcall permitted only in pattern actions "
+						"of a longest match" << endl;
+					break;
 				}
+				check = check->parent;
 			}
 		}
 	}
@@ -1635,6 +1277,11 @@ void ParseData::analyzeGraph( FsmAp *graph )
 
 		for ( ActionTable::Iter at = st->eofActionTable; at.lte(); at++ )
 			at->value->numEofRefs += 1;
+
+		for ( StateCondList::Iter sc = st->stateCondList; sc.lte(); sc++ ) {
+			for ( CondSet::Iter sci = sc->condSpace->condSet; sci.lte(); sci++ )
+				(*sci)->numCondRefs += 1;
+		}
 	}
 
 	/* Checks for bad usage of directives in action code. */
@@ -1645,7 +1292,6 @@ void ParseData::analyzeGraph( FsmAp *graph )
 void ParseData::prepareMachineGen( GraphDictEl *graphDictEl )
 {
 	beginProcessing();
-	setLowerUpperRange();
 	initKeyOps();
 	makeRootName();
 	initLongestMatchData();
@@ -1675,6 +1321,12 @@ void ParseData::generateXML( ostream &out )
 
 	/* Write out with it. */
 	codeGen.writeXML();
+
+	if ( printStatistics ) {
+		cerr << "fsm name  : " << fsmName << endl;
+		cerr << "num states: " << sectionGraph->stateList.length() << endl;
+		cerr << endl;
+	}
 }
 
 void checkMachines( )
@@ -1693,6 +1345,18 @@ void checkMachines( )
 	}
 }
 
+void writeLanguage( std::ostream &out )
+{
+	out << " lang=\"";
+	switch ( hostLangType ) {
+		case CCode:    out << "C"; break;
+		case DCode:    out << "D"; break;
+		case JavaCode: out << "Java"; break;
+	}
+	out << "\"";
+	
+}
+
 void writeMachines( std::ostream &out, std::string hostData, char *inputFileName )
 {
 	if ( machineSpec == 0 && machineName == 0 ) {
@@ -1704,7 +1368,9 @@ void writeMachines( std::ostream &out, std::string hostData, char *inputFileName
 		}
 
 		if ( gblErrorCount == 0 ) {
-			out << "<ragel filename=\"" << inputFileName << "\">\n";
+			out << "<ragel filename=\"" << inputFileName << "\"";
+			writeLanguage( out );
+			out << ">\n";
 			for ( SectionMap::Iter sect = sectionMap; sect.lte(); sect++ ) {
 				ParseData *pd = sect->value;
 				if ( pd->instanceList.length() > 0 )
@@ -1743,7 +1409,9 @@ void writeMachines( std::ostream &out, std::string hostData, char *inputFileName
 			/* Section/Machine to emit was found. Prepare and emit it. */
 			parseData->prepareMachineGen( graphDictEl );
 			if ( gblErrorCount == 0 ) {
-				out << "<ragel filename=\"" << inputFileName << "\">\n";
+				out << "<ragel filename=\"" << inputFileName << "\"";
+				writeLanguage( out );
+				out << ">\n";
 				parseData->generateXML( out );
 				out << hostData;
 				out << "</ragel>\n";
