@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  */
 
 /*  This file is part of Ragel.
@@ -22,6 +22,7 @@
 #ifndef _PARSEDATA_H
 #define _PARSEDATA_H
 
+#include <iostream>
 #include <limits.h>
 #include "avlmap.h"
 #include "bstmap.h"
@@ -58,7 +59,7 @@ enum BuiltinMachine
 	BT_Empty
 };
 
-struct BISON_YYLTYPE;
+struct YYLTYPE;
 
 /* Location in an input file. */
 struct InputLoc
@@ -67,7 +68,7 @@ struct InputLoc
 		: line(0), col(0) { }
 	InputLoc( int line, int col ) 
 		: line(line), col(col) { }
-	InputLoc( const BISON_YYLTYPE &loc );
+	InputLoc( const YYLTYPE &loc );
 
 	int line;
 	int col;
@@ -100,6 +101,7 @@ public:
 		numToStateRefs(0),
 		numFromStateRefs(0),
 		numEofRefs(0),
+		numCondRefs(0),
 		anyCall(false),
 		isLmAction(false),
 		nameRefs(nameRefs)
@@ -115,6 +117,14 @@ public:
 	InlineList *inlineList;
 	int actionId;
 
+	void actionName( ostream &out )
+	{
+		if ( name != 0 )
+			out << name;
+		else
+			out << loc.line << ":" << loc.col;
+	}
+
 	/* Places in the input text that reference the action. */
 	ActionRefs actionRefs;
 
@@ -125,6 +135,7 @@ public:
 	int numToStateRefs;
 	int numFromStateRefs;
 	int numEofRefs;
+	int numCondRefs;
 	bool anyCall;
 
 	bool isLmAction;
@@ -160,9 +171,10 @@ struct PriorityAug
  * factor with aug will keep an array of these. */
 struct ParserAction
 {
-	ParserAction( AugType type, int localErrKey, Action *action )
-		: type(type), localErrKey(localErrKey), action(action) { }
+	ParserAction( const InputLoc &loc, AugType type, int localErrKey, Action *action )
+		: loc(loc), type(type), localErrKey(localErrKey), action(action) { }
 
+	InputLoc loc;
 	AugType type;
 	int localErrKey;
 	Action *action;
@@ -343,17 +355,8 @@ struct ParseData
 	FsmAp *sectionGraph;
 	bool generatingSectionSubset;
 
-	/* Set the lower and upper range from the lower and upper number keys. */
-	void setLowerUpperRange( );
 	void initKeyOps();
 
-	/*
-	 * Querying the parse data
-	 */
-
-	/* Is the alphabet a signed type? */
-	bool isAlphSigned();
-	
 	/*
 	 * Data collected during the parse.
 	 */
@@ -386,7 +389,7 @@ struct ParseData
 	int curDefLocalErrKey;
 
 	/* Alphabet type. */
-	AlphType alphType;
+	HostType *userAlphType;
 	bool alphTypeSet;
 
 	/* Element type and get key expression. */
@@ -396,7 +399,7 @@ struct ParseData
 
 	/* The alphabet range. */
 	char *lowerNum, *upperNum;
-	long lowKey, highKey;
+	Key lowKey, highKey;
 	InputLoc rangeLowLoc, rangeHighLoc;
 
 	/* The name of the file the fsm is from, and the spec name. */
@@ -472,13 +475,13 @@ typedef AvlMap<char*, ParseData *, CmpStr> SectionMap;
 typedef AvlMapEl<char*, ParseData *> SectionMapEl;
 extern SectionMap sectionMap;
 
-void afterOpMinimize( FsmAp *fsm );
-long makeFsmKeyHex( char *str, const InputLoc &loc, ParseData *pd );
-long makeFsmKeyDec( char *str, const InputLoc &loc, ParseData *pd );
-long makeFsmKeyNum( char *str, const InputLoc &loc, ParseData *pd );
-long makeFsmKeyChar( char c, ParseData *pd );
-void makeFsmKeyArray( long *result, char *data, int len, ParseData *pd );
-long *makeFsmUniqueKeyArray( int &resLen, char *data, int len, 
+void afterOpMinimize( FsmAp *fsm, bool lastInSeq = true );
+Key makeFsmKeyHex( char *str, const InputLoc &loc, ParseData *pd );
+Key makeFsmKeyDec( char *str, const InputLoc &loc, ParseData *pd );
+Key makeFsmKeyNum( char *str, const InputLoc &loc, ParseData *pd );
+Key makeFsmKeyChar( char c, ParseData *pd );
+void makeFsmKeyArray( Key *result, char *data, int len, ParseData *pd );
+void makeFsmUniqueKeyArray( KeySet &result, char *data, int len, 
 		bool caseInsensitive, ParseData *pd );
 FsmAp *makeBuiltin( BuiltinMachine builtin, ParseData *pd );
 FsmAp *dotFsm( ParseData *pd );
