@@ -202,7 +202,7 @@ struct Expression
 	~Expression();
 
 	/* Tree traversal. */
-	FsmAp *walk( ParseData *pd );
+	FsmAp *walk( ParseData *pd, bool lastInSeq = true );
 	void makeNameTree( ParseData *pd );
 	void resolveNameRefs( ParseData *pd );
 
@@ -239,7 +239,7 @@ struct Term
 	
 	~Term();
 
-	FsmAp *walk( ParseData *pd );
+	FsmAp *walk( ParseData *pd, bool lastInSeq = true );
 	void makeNameTree( ParseData *pd );
 	void resolveNameRefs( ParseData *pd );
 
@@ -285,12 +285,15 @@ struct FactorWithAug
 	void assignActions( ParseData *pd, FsmAp *graph, int *actionOrd );
 	void assignPriorities( FsmAp *graph, int *priorOrd );
 
+	void assignConditions( FsmAp *graph );
+
 	/* Actions and priorities assigned to the factor node. */
 	Vector<ParserAction> actions;
 	Vector<PriorityAug> priorityAugs;
 	PriorDesc *priorDescs;
 	Vector<Label> labels;
 	Vector<EpsilonLink> epsilonLinks;
+	Vector<ParserAction> conditions;
 
 	FactorWithRep *factorWithRep;
 };
@@ -426,7 +429,6 @@ struct Factor
 	Join *join;
 	LongestMatch *longestMatch;
 	int lower, upper;
-	Vector<Action*> conditions;
 	Type type;
 };
 
@@ -451,7 +453,7 @@ struct Literal
 	Literal( const InputLoc &loc, TokenData data, LiteralType type )
 		: loc(loc), data(data), type(type), caseInsensitive(false) { }
 
-	long makeRangeEndPoint( ParseData *pd );
+	Key makeRangeEndPoint( ParseData *pd );
 	FsmAp *walk( ParseData *pd );
 	
 	InputLoc loc;
