@@ -1,5 +1,5 @@
 /*
- *  Copyright 2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2005, 2006 Adrian Thurston <thurston@cs.queensu.ca>
  */
 
 /*  This file is part of Ragel.
@@ -44,7 +44,7 @@ void XMLCodeGen::writeActionList()
 	/* Determine which actions to write. */
 	int nextActionId = 0;
 	for ( ActionList::Iter act = pd->actionList; act.lte(); act++ ) {
-		if ( ! act->isLmAction || act->numRefs() > 0 )
+		if ( act->numRefs() > 0 || act->numCondRefs > 0 )
 			act->actionId = nextActionId++;
 	}
 
@@ -115,22 +115,22 @@ void XMLCodeGen::reduceActionTables()
 	}
 }
 
-void XMLCodeGen::appendTrans( TransListVect &outList, long lowKey, 
-		long highKey, TransAp *trans )
+void XMLCodeGen::appendTrans( TransListVect &outList, Key lowKey, 
+		Key highKey, TransAp *trans )
 {
 	if ( trans->toState != 0 || trans->actionTable.length() > 0 )
 		outList.append( TransEl( lowKey, highKey, trans ) );
 }
 
-void XMLCodeGen::writeKey( long key )
+void XMLCodeGen::writeKey( Key key )
 {
-	if ( keyOps->isAlphSigned )
-		out << key;
+	if ( keyOps->isSigned )
+		out << key.getVal();
 	else
-		out << (unsigned long) key;
+		out << (unsigned long) key.getVal();
 }
 
-void XMLCodeGen::writeTrans( long lowKey, long highKey, TransAp *trans )
+void XMLCodeGen::writeTrans( Key lowKey, Key highKey, TransAp *trans )
 {
 	/* First reduce the action. */
 	RedActionTable *actionTable = 0;
@@ -478,16 +478,18 @@ void XMLCodeGen::writeStateActions( StateAp *state )
 
 void XMLCodeGen::writeStateConditions( StateAp *state )
 {
-	if ( state->stateCondDict.length() > 0 ) {
-		out << "      <cond>";
-		for ( StateCondDict::Iter scdi = state->stateCondDict; scdi.lte(); scdi++ ) {
-			Condition *condition = condData->condDict.find( scdi->key );
-			CondSetMapEl *csMapEl = condition->condSetMap.find( scdi->condSet );
-			assert( condition != 0 );
-			assert( csMapEl != 0 );
-			out << csMapEl->conditionId;
+	if ( state->stateCondList.length() > 0 ) {
+		out << "      <cond_list length=\"" << state->stateCondList.length() << "\">\n";
+		for ( StateCondList::Iter scdi = state->stateCondList; scdi.lte(); scdi++ ) {
+			out << "        <c>";
+			writeKey( scdi->lowKey );
+			out << " ";
+			writeKey( scdi->highKey );
+			out << " ";
+			out << scdi->condSpace->condSpaceId;
+			out << "</c>\n";
 		}
-		out << "</cond>\n";
+		out << "      </cond_list>\n";
 	}
 }
 
@@ -559,34 +561,8 @@ void XMLCodeGen::writeMachine()
 
 void XMLCodeGen::writeAlphType()
 {
-	out << "  <alphtype>";
-	switch ( pd->alphType ) {
-		case AT_Char:
-			out << "signed_char";
-			break;
-		case AT_UnsignedChar:
-			out << "unsigned_char";
-			break;
-		case AT_Short:
-			out << "signed_short";
-			break;
-		case AT_UnsignedShort:
-			out << "unsigned_short";
-			break;
-		case AT_Int:
-			out << "signed_int";
-			break;
-		case AT_UnsignedInt:
-			out << "unsigned_int";
-			break;
-		case AT_Long:
-			out << "signed_long";
-			break;
-		case AT_UnsignedLong:
-			out << "unsigned_long";
-			break;
-	}
-	out << "</alphtype>\n";
+	out << "  <alphtype>" << 
+		(keyOps->alphType - hostLang->hostTypes) << "</alphtype>\n";
 }
 
 void XMLCodeGen::writeGetKeyExpr()
@@ -612,34 +588,21 @@ void XMLCodeGen::writeCurStateExpr()
 
 void XMLCodeGen::writeConditions()
 {
-	int nextCondId = 0;
-	for ( CondDict::Iter cond = condData->condDict; cond.lte(); cond++ ) {
-		for ( CondSetMap::Iter mapItem = cond->condSetMap; mapItem.lte(); mapItem++ )
-			mapItem->conditionId = nextCondId++;
-	}
+	if ( condData->condSpaceMap.length() > 0 ) {
+		long nextCondSpaceId = 0;
+		for ( CondSpaceMap::Iter cs = condData->condSpaceMap; cs.lte(); cs++ )
+			cs->condSpaceId = nextCondSpaceId++;
 
-	if ( condData->condDict.length() > 0 ) {
-		int length = 0;
-		for ( CondDict::Iter cond = condData->condDict; cond.lte(); cond++ )
-			length += cond->condSetMap.length();
-
-		out << "    <cond_list length=\"" << length << "\">\n";
-		for ( CondDict::Iter cond = condData->condDict; cond.lte(); cond++ ) {
-			for ( CondSetMap::Iter mapItem = cond->condSetMap; mapItem.lte(); mapItem++ ) {
-				CondSet &condSet = mapItem->condSet;
-
-				out << "      <cond id=\"" << 
-					"\" length=\"" << condSet.length() << "\">";
-				writeKey( cond->key );
-				out << " ";
-				writeKey( mapItem->baseKey );
-
-				for ( CondSet::Iter csi = condSet; csi.lte(); csi++ )
-					out << " " << (*csi)->actionId;
-				out << "</cond>\n";
-			}
+		out << "    <cond_space_list length=\"" << condData->condSpaceMap.length() << "\">\n";
+		for ( CondSpaceMap::Iter cs = condData->condSpaceMap; cs.lte(); cs++ ) {
+			out << "      <cond_space id=\"" << cs->condSpaceId << 
+				"\" length=\"" << cs->condSet.length() << "\">";
+			writeKey( cs->baseKey );
+			for ( CondSet::Iter csi = cs->condSet; csi.lte(); csi++ )
+				out << " " << (*csi)->actionId;
+			out << "</cond_space>\n";
 		}
-		out << "    </cond_list>\n";
+		out << "    </cond_space_list>\n";
 	}
 }
 
@@ -647,9 +610,7 @@ void XMLCodeGen::writeXML()
 {
 	/* Open the definition. */
 	out << "<ragel_def name=\"" << fsmName << "\">\n";
-
-	if ( pd->alphTypeSet )
-		writeAlphType();
+	writeAlphType();
 	
 	if ( pd->getKeyExpr != 0 )
 		writeGetKeyExpr();
