@@ -1,5 +1,5 @@
 /*
- *  Copyright 2004, 2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2004-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -151,6 +151,28 @@ std::ostream &FFlatCodeGen::ACTION_SWITCH()
 
 void FFlatCodeGen::writeOutData()
 {
+	if ( anyConditions() ) {
+		OPEN_ARRAY( WIDE_ALPH_TYPE(), CK() );
+		COND_KEYS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( ARRAY_TYPE(maxCondSpan), CSP() );
+		COND_KEY_SPANS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( ARRAY_TYPE(maxCond), C() );
+		CONDS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( ARRAY_TYPE(maxCondIndexOffset), CO() );
+		COND_INDEX_OFFSET();
+		CLOSE_ARRAY() <<
+		"\n";
+	}
+
 	OPEN_ARRAY( WIDE_ALPH_TYPE(), K() );
 	KEYS();
 	CLOSE_ARRAY() <<
@@ -205,68 +227,66 @@ void FFlatCodeGen::writeOutData()
 	}
 
 	out <<
-		"static int " << START() << " = " << START_STATE_ID() << ";\n"
+		"static const int " << START() << " = " << START_STATE_ID() << ";\n"
 		"\n";
 
 	if ( cgd->writeFirstFinal ) {
 		out <<
-			"static int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
+			"static const int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
 			"\n";
 	}
 
 	if ( cgd->writeErr ) {
 		out <<
-			"static int " << ERROR() << " = " << ERROR_STATE() << ";\n"
+			"static const int " << ERROR() << " = " << ERROR_STATE() << ";\n"
 			"\n";
 	}
 }
 
 void FFlatCodeGen::writeOutExec()
 {
+	outLabelUsed = false;
+
 	out << 
-		"		{\n"
-		"		int _slen";
+		"	{\n"
+		"	int _slen";
 
 	if ( anyRegCurStateRef() )
 		out << ", _ps";
 	
-	out <<
-		";\n"
-		"		" << WIDE_ALPH_TYPE() << POINTER() << "_keys;\n"
-		"		" << ARRAY_TYPE(maxIndex) << POINTER() << "_inds;\n"
-		"		int _trans;\n";
+	out << ";\n";
+	out << "	int _trans";
 
 	if ( anyConditions() )
-		out << "	" << WIDE_ALPH_TYPE() << " _widec;\n";
+		out << ", _cond";
 
-	out << 
-		"	" << P() << " -= 1;\n"
-		"	goto _resume;\n"
-		"\n";
+	out << ";\n";
 
-	/* Head of the loop. */
-	out << "_again:\n";
+	out <<
+		"	" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_keys;\n"
+		"	" << PTR_CONST() << ARRAY_TYPE(maxIndex) << POINTER() << "_inds;\n";
 
-	if ( anyToStateActions() ) {
-		out <<
-			"	switch ( " << TSA() << "[" << CS() << "] ) {\n";
-			TO_STATE_ACTION_SWITCH();
-			SWITCH_DEFAULT() <<
-			"	}\n"
-			"\n";
+	if ( anyConditions() ) {
+		out << 
+			"	" << PTR_CONST() << ARRAY_TYPE(maxCond) << POINTER() << "_conds;\n"
+			"	" << WIDE_ALPH_TYPE() << " _widec;\n";
+	}
+
+	if ( cgd->hasEnd ) {
+		outLabelUsed = true;
+		out << 
+			"	if ( " << P() << " == " << PE() << " )\n"
+			"		goto _out;\n";
 	}
 
 	out << "_resume:\n";
 
-	if ( cgd->hasEnd ) {
+	if ( redFsm->errState != 0 ) {
+		outLabelUsed = true;
 		out << 
-			"	if ( ++" << P() << " == " << PE() << " )\n"
+			"	if ( " << CS() << " == " << redFsm->errState->id << " )\n"
 			"		goto _out;\n";
 	}
-	else {
-		out << 
-			"	" << P() << " += 1;\n";
-	}
 
 	if ( anyFromStateActions() ) {
 		out <<
@@ -282,12 +302,6 @@ void FFlatCodeGen::writeOutExec()
 
 	LOCATE_TRANS();
 	
-	if ( redFsm->errState != 0 ) {
-		out << 
-			"	if ( " << CS() << " == " << redFsm->errState->id << " )\n"
-			"		goto _out;\n";
-	}
-
 	if ( anyRegCurStateRef() )
 		out << "	_ps = " << CS() << ";\n";
 
@@ -306,9 +320,30 @@ void FFlatCodeGen::writeOutExec()
 			"\n";
 	}
 
-	out << "	goto _again;\n";
+	if ( anyRegActions() || anyActionGotos() || anyActionCalls() || anyActionRets() )
+		out << "_again:\n";
 
-	if ( cgd->hasEnd || anyRegBreak() || redFsm->errState != 0 )
+	if ( anyToStateActions() ) {
+		out <<
+			"	switch ( " << TSA() << "[" << CS() << "] ) {\n";
+			TO_STATE_ACTION_SWITCH();
+			SWITCH_DEFAULT() <<
+			"	}\n"
+			"\n";
+	}
+
+	if ( cgd->hasEnd ) {
+		out << 
+			"	if ( ++" << P() << " != " << PE() << " )\n"
+			"		goto _resume;\n";
+	}
+	else {
+		out << 
+			"	" << P() << " += 1;\n"
+			"	goto _resume;\n";
+	}
+
+	if ( outLabelUsed )
 		out << "	_out: {}\n";
 
 	out << "	}\n";
