@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -153,18 +153,18 @@ unsigned int FGotoCodeGen::EOF_ACTION( RedStateAp *state )
 void FGotoCodeGen::writeOutData()
 {
 	out <<
-		"static int " << START() << " = " << START_STATE_ID() << ";\n"
+		"static const int " << START() << " = " << START_STATE_ID() << ";\n"
 		"\n";
 
 	if ( cgd->writeFirstFinal ) {
 		out <<
-			"static int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
+			"static const int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
 			"\n";
 	}
 
 	if ( cgd->writeErr ) {
 		out <<
-			"static int " << ERROR() << " = " << ERROR_STATE() << ";\n"
+			"static const int " << ERROR() << " = " << ERROR_STATE() << ";\n"
 			"\n";
 	}
 
@@ -192,6 +192,8 @@ void FGotoCodeGen::writeOutData()
 
 void FGotoCodeGen::writeOutExec()
 {
+	outLabelUsed = false;
+
 	out << "	{\n";
 
 	if ( anyRegCurStateRef() )
@@ -200,35 +202,15 @@ void FGotoCodeGen::writeOutExec()
 	if ( anyConditions() )
 		out << "	" << WIDE_ALPH_TYPE() << " _widec;\n";
 
-	out << 
-		"	" << P() << " -= 1;\n"
-		"	goto _resume;\n"
-		"\n";
-
-	/* Head of the loop. */
-	out << "_again:\n";
-
-	if ( anyToStateActions() ) {
-		out <<
-			"	switch ( " << TSA() << "[" << CS() << "] ) {\n";
-			TO_STATE_ACTION_SWITCH();
-			SWITCH_DEFAULT() <<
-			"	}\n"
-			"\n";
+	if ( cgd->hasEnd ) {
+		outLabelUsed = true;
+		out << 
+			"	if ( " << P() << " == " << PE() << " )\n"
+			"		goto _out;\n";
 	}
 
 	out << "_resume:\n";
 
-	if ( cgd->hasEnd ) {
-		out << 
-			"	if ( ++" << P() << " == " << PE() << " )\n"
-			"		goto _out;\n";
-	}
-	else {
-		out << 
-			"	" << P() << " += 1;\n";
-	}
-
 	if ( anyFromStateActions() ) {
 		out <<
 			"	switch ( " << FSA() << "[" << CS() << "] ) {\n";
@@ -250,7 +232,30 @@ void FGotoCodeGen::writeOutExec()
 	if ( anyRegActions() )
 		EXEC_ACTIONS() << "\n";
 
-	if ( cgd->hasEnd || anyRegBreak() || redFsm->errState != 0 )
+	out << "_again:\n";
+
+	if ( anyToStateActions() ) {
+		out <<
+			"	switch ( " << TSA() << "[" << CS() << "] ) {\n";
+			TO_STATE_ACTION_SWITCH();
+			SWITCH_DEFAULT() <<
+			"	}\n"
+			"\n";
+	}
+
+	if ( cgd->hasEnd ) {
+		out << 
+			"	if ( ++" << P() << " != " << PE() << " )\n"
+			"		goto _resume;\n";
+	}
+	else {
+		out << 
+			"	" << P() << " += 1;\n"
+			"	goto _resume;\n";
+	}
+
+
+	if ( outLabelUsed )
 		out << "	_out: {}\n";
 
 	out << "	}\n";
