@@ -1,5 +1,5 @@
 /*
- *  Copyright 2004 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2004-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -148,7 +148,7 @@ std::ostream &FlatCodeGen::FLAT_INDEX_OFFSET()
 		
 		/* Move the index offset ahead. */
 		if ( st->transList != 0 )
-			curIndOffset += redFsm->keyOps->span( st->lowKey, st->highKey );
+			curIndOffset += keyOps->span( st->lowKey, st->highKey );
 
 		if ( st->defTrans != 0 )
 			curIndOffset += 1;
@@ -165,7 +165,7 @@ std::ostream &FlatCodeGen::KEY_SPANS()
 		/* Write singles length. */
 		unsigned long long span = 0;
 		if ( st->transList != 0 )
-			span = redFsm->keyOps->span( st->lowKey, st->highKey );
+			span = keyOps->span( st->lowKey, st->highKey );
 		out << span;
 		if ( !st.last() ) {
 			out << ", ";
@@ -228,6 +228,90 @@ std::ostream &FlatCodeGen::EOF_ACTIONS()
 	return out;
 }
 
+std::ostream &FlatCodeGen::COND_KEYS()
+{
+	out << '\t';
+	int totalTrans = 0;
+	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Emit just cond low key and cond high key. */
+		out << KEY( st->condLowKey ) << ", ";
+		out << KEY( st->condHighKey ) << ", ";
+		if ( ++totalTrans % IALL == 0 )
+			out << "\n\t";
+	}
+
+	/* Output one last number so we don't have to figure out when the last
+	 * entry is and avoid writing a comma. */
+	out << 0 << "\n";
+	return out;
+}
+
+std::ostream &FlatCodeGen::COND_KEY_SPANS()
+{
+	out << "\t";
+	int totalStateNum = 0;
+	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Write singles length. */
+		unsigned long long span = 0;
+		if ( st->condList != 0 )
+			span = keyOps->span( st->condLowKey, st->condHighKey );
+		out << span;
+		if ( !st.last() ) {
+			out << ", ";
+			if ( ++totalStateNum % IALL == 0 )
+				out << "\n\t";
+		}
+	}
+	out << "\n";
+	return out;
+}
+
+std::ostream &FlatCodeGen::CONDS()
+{
+	int totalTrans = 0;
+	out << '\t';
+	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		if ( st->condList != 0 ) {
+			/* Walk the singles. */
+			unsigned long long span = keyOps->span( st->condLowKey, st->condHighKey );
+			for ( unsigned long long pos = 0; pos < span; pos++ ) {
+				if ( st->condList[pos] != 0 )
+					out << st->condList[pos]->condSpaceId + 1 << ", ";
+				else
+					out << "0, ";
+				if ( ++totalTrans % IALL == 0 )
+					out << "\n\t";
+			}
+		}
+	}
+
+	/* Output one last number so we don't have to figure out when the last
+	 * entry is and avoid writing a comma. */
+	out << 0 << "\n";
+	return out;
+}
+
+std::ostream &FlatCodeGen::COND_INDEX_OFFSET()
+{
+	out << "\t";
+	int totalStateNum = 0, curIndOffset = 0;
+	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Write the index offset. */
+		out << curIndOffset;
+		if ( !st.last() ) {
+			out << ", ";
+			if ( ++totalStateNum % IALL == 0 )
+				out << "\n\t";
+		}
+		
+		/* Move the index offset ahead. */
+		if ( st->condList != 0 )
+			curIndOffset += keyOps->span( st->condLowKey, st->condHighKey );
+	}
+	out << "\n";
+	return out;
+}
+
 
 std::ostream &FlatCodeGen::KEYS()
 {
@@ -254,7 +338,7 @@ std::ostream &FlatCodeGen::INDICIES()
 	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
 		if ( st->transList != 0 ) {
 			/* Walk the singles. */
-			unsigned long long span = redFsm->keyOps->span( st->lowKey, st->highKey );
+			unsigned long long span = keyOps->span( st->lowKey, st->highKey );
 			for ( unsigned long long pos = 0; pos < span; pos++ ) {
 				out << st->transList[pos]->id << ", ";
 				if ( ++totalTrans % IALL == 0 )
@@ -396,6 +480,7 @@ void FlatCodeGen::RET( ostream &ret, bool inFinish )
 
 void FlatCodeGen::BREAK( ostream &ret, int targState )
 {
+	outLabelUsed = true;
 	ret << "goto _out;";
 }
 
@@ -410,6 +495,28 @@ void FlatCodeGen::writeOutData()
 		"\n";
 	}
 
+	if ( anyConditions() ) {
+		OPEN_ARRAY( WIDE_ALPH_TYPE(), CK() );
+		COND_KEYS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( ARRAY_TYPE(maxCondSpan), CSP() );
+		COND_KEY_SPANS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( ARRAY_TYPE(maxCond), C() );
+		CONDS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( ARRAY_TYPE(maxCondIndexOffset), CO() );
+		COND_INDEX_OFFSET();
+		CLOSE_ARRAY() <<
+		"\n";
+	}
+
 	OPEN_ARRAY( WIDE_ALPH_TYPE(), K() );
 	KEYS();
 	CLOSE_ARRAY() <<
@@ -464,24 +571,64 @@ void FlatCodeGen::writeOutData()
 	}
 
 	out <<
-		"static int " << START() << " = " << START_STATE_ID() << ";\n"
+		"static const int " << START() << " = " << START_STATE_ID() << ";\n"
 		"\n";
 
 	if ( cgd->writeFirstFinal ) {
 		out <<
-			"static int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
+			"static const int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
 			"\n";
 	}
 
 	if ( cgd->writeErr ) {
 		out <<
-			"static int " << ERROR() << " = " << ERROR_STATE() << ";\n"
+			"static const int " << ERROR() << " = " << ERROR_STATE() << ";\n"
 			"\n";
 	}
 }
 
+void FlatCodeGen::COND_TRANSLATE()
+{
+	out << 
+		"	_widec = " << GET_KEY() << ";\n";
+
+	out <<
+		"	_keys = " << ARR_OFF( CK(), "(" + CS() + "<<1)" ) << ";\n"
+		"	_conds = " << ARR_OFF( C(), CO() + "[" + CS() + "]" ) << ";\n"
+		"\n"
+		"	_slen = " << CSP() << "[" << CS() << "];\n"
+		"	_cond = _slen > 0 && _keys[0] <=" << GET_WIDE_KEY() << " &&\n"
+		"		" << GET_WIDE_KEY() << " <= _keys[1] ?\n"
+		"		_conds[" << GET_WIDE_KEY() << " - _keys[0]] : 0;\n"
+		"\n";
+
+	out <<
+		"	switch ( _cond ) {\n";
+	for ( CondSpaceList::Iter csi = cgd->condSpaceList; csi.lte(); csi++ ) {
+		CondSpace *condSpace = csi;
+		out << "	case " << condSpace->condSpaceId + 1 << ": {\n";
+		out << TABS(2) << "_widec = " << KEY(condSpace->baseKey) << 
+				" + (" << GET_KEY() << " - " << KEY(keyOps->minKey) << ");\n";
+
+		for ( CondSet::Iter csi = condSpace->condSet; csi.lte(); csi++ ) {
+			out << TABS(2) << "if ( ";
+			CONDITION( out, *csi );
+			Size condValOffset = ((1 << csi.pos()) * keyOps->alphSize());
+			out << " ) _widec += " << condValOffset << ";\n";
+		}
+
+		out << "		}\n";
+		out << "		break;\n";
+	}
+
+	out <<
+		"	}\n";
+}
+
 void FlatCodeGen::writeOutExec()
 {
+	outLabelUsed = false;
+
 	out << 
 		"	{\n"
 		"	int _slen";
@@ -491,54 +638,44 @@ void FlatCodeGen::writeOutExec()
 
 	out << 
 		";\n"
-		"	int _trans;\n";
+		"	int _trans";
 
-	if ( anyActions() ) {
+	if ( anyConditions() )
+		out << ", _cond";
+	out << ";\n";
+
+	if ( anyToStateActions() || anyRegActions() || anyFromStateActions() ) {
 		out << 
-			"	" << ARRAY_TYPE(maxActArrItem) << POINTER() << "_acts;\n"
+			"	" << PTR_CONST() << ARRAY_TYPE(maxActArrItem) << POINTER() << "_acts;\n"
 			"	" << UINT() << " _nacts;\n"; 
 	}
 
 	out <<
-		"	" << WIDE_ALPH_TYPE() << POINTER() << "_keys;\n"
-		"	" << ARRAY_TYPE(maxIndex) << POINTER() << "_inds;\n";
+		"	" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_keys;\n"
+		"	" << PTR_CONST() << ARRAY_TYPE(maxIndex) << POINTER() << "_inds;\n";
 
-	if ( anyConditions() )
-		out << "	" << WIDE_ALPH_TYPE() << " _widec;\n";
+	if ( anyConditions() ) {
+		out << 
+			"	" << PTR_CONST() << ARRAY_TYPE(maxCond) << POINTER() << "_conds;\n"
+			"	" << WIDE_ALPH_TYPE() << " _widec;\n";
+	}
 
 	out << "\n";
-	
-	out << 
-		"	" << P() << " -= 1;\n"
-		"	goto _resume;\n"
-		"\n";
-
-	/* Head of the loop. */
-	out << "_again:\n";
-
-	if ( anyToStateActions() ) {
-		out <<
-			"	_acts = " << ARR_OFF( A(),  TSA() + "[" + CS() + "]" ) << ";\n"
-			"	_nacts = " << CAST(UINT()) << " *_acts++;\n"
-			"	while ( _nacts-- > 0 ) {\n"
-			"		switch ( *_acts++ ) {\n";
-			TO_STATE_ACTION_SWITCH();
-			SWITCH_DEFAULT() <<
-			"		}\n"
-			"	}\n"
-			"\n";
-	}
-
-	out << "_resume:\n";
 
 	if ( cgd->hasEnd ) {
+		outLabelUsed = true;
 		out << 
-			"	if ( ++" << P() << " == " << PE() << " )\n"
+			"	if ( " << P() << " == " << PE() << " )\n"
 			"		goto _out;\n";
 	}
-	else {
+	
+	out << "_resume:\n";
+
+	if ( redFsm->errState != 0 ) {
+		outLabelUsed = true;
 		out << 
-			"	" << P() << " += 1;\n";
+			"	if ( " << CS() << " == " << redFsm->errState->id << " )\n"
+			"		goto _out;\n";
 	}
 
 	if ( anyFromStateActions() ) {
@@ -559,12 +696,6 @@ void FlatCodeGen::writeOutExec()
 
 	LOCATE_TRANS();
 
-	if ( redFsm->errState != 0 ) {
-		out << 
-			"	if ( " << CS() << " == " << redFsm->errState->id << " )\n"
-			"		goto _out;\n";
-	}
-
 	if ( anyRegCurStateRef() )
 		out << "	_ps = " << CS() << ";\n";
 
@@ -588,9 +719,34 @@ void FlatCodeGen::writeOutExec()
 			"\n";
 	}
 
-	out << "	goto _again;\n";
-	
-	if ( cgd->hasEnd || anyRegBreak() || redFsm->errState != 0 )
+	if ( anyRegActions() || anyActionGotos() || anyActionCalls() || anyActionRets() )
+		out << "_again:\n";
+
+	if ( anyToStateActions() ) {
+		out <<
+			"	_acts = " << ARR_OFF( A(),  TSA() + "[" + CS() + "]" ) << ";\n"
+			"	_nacts = " << CAST(UINT()) << " *_acts++;\n"
+			"	while ( _nacts-- > 0 ) {\n"
+			"		switch ( *_acts++ ) {\n";
+			TO_STATE_ACTION_SWITCH();
+			SWITCH_DEFAULT() <<
+			"		}\n"
+			"	}\n"
+			"\n";
+	}
+
+	if ( cgd->hasEnd ) {
+		out << 
+			"	if ( ++" << P() << " != " << PE() << " )\n"
+			"		goto _resume;\n";
+	}
+	else {
+		out << 
+			"	" << P() << " += 1;\n"
+			"	goto _resume;\n";
+	}
+
+	if ( outLabelUsed )
 		out << "	_out: {}\n";
 
 	out << "	}\n";
@@ -601,7 +757,7 @@ void FlatCodeGen::writeOutEOF()
 	if ( anyEofActions() ) {
 		out << 
 			"	{\n"
-			"	" << ARRAY_TYPE(maxActArrItem) << " *_acts = " << 
+			"	" << PTR_CONST() << ARRAY_TYPE(maxActArrItem) << POINTER() << "_acts = " << 
 					ARR_OFF( A(), EA() + "[" + CS() + "]" ) << ";\n"
 			"	" << UINT() << " _nacts = " << CAST(UINT()) << " *_acts++;\n"
 			"	while ( _nacts-- > 0 ) {\n"
