@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -26,6 +26,15 @@
 #include "redfsm.h"
 #include "gendata.h"
 #include <sstream>
+#include <string>
+#include <assert.h>
+
+using std::ostream;
+using std::ostringstream;
+using std::string;
+using std::cerr;
+using std::endl;
+
 
 /* Determine if a string is only whitespace. Code blocks that are only
  * whitespace need not be output. */
@@ -40,9 +49,6 @@ bool onlyWhitespace( char *str )
 	return true;
 }
 
-using std::ostream;
-using std::ostringstream;
-
 /* Init code gen with in parameters. */
 FsmCodeGen::FsmCodeGen( )
 :
@@ -54,6 +60,7 @@ FsmCodeGen::FsmCodeGen( )
 	bAnyFromStateActions(false),
 	bAnyRegActions(false),
 	bAnyEofActions(false),
+	bAnyActionGotos(false),
 	bAnyActionCalls(false),
 	bAnyActionRets(false),
 	bAnyRegActionRets(false),
@@ -66,11 +73,9 @@ FsmCodeGen::FsmCodeGen( )
 {
 }
 
-/* Does the machine have any functions. */
+/* Does the machine have any actions. */
 bool FsmCodeGen::anyActions()
 {
-	/* If there is more than one action then one must be non-empty. If there
-	 * is only one then it may be nonzero. */
 	return redFsm->actionMap.length() > 0;
 }
 
@@ -142,17 +147,35 @@ void FsmCodeGen::setValueLimits()
 	maxRangeLen = 0;
 	maxKeyOffset = 0;
 	maxIndexOffset = 0;
-	maxIndex = redFsm->transSet.length();
 	maxActListId = 0;
 	maxActionLoc = 0;
 	maxActArrItem = 0;
 	maxSpan = 0;
+	maxCondSpan = 0;
 	maxFlatIndexOffset = 0;
+	maxCondOffset = 0;
+	maxCondLen = 0;
+	maxCondSpaceId = 0;
+	maxCondIndexOffset = 0;
+
+	/* In both of these cases the 0 index is reserved for no value, so the max
+	 * is one more than it would be if they started at 0. */
+	maxIndex = redFsm->transSet.length();
+	maxCond = cgd->condSpaceList.length(); 
 
 	/* The nextStateId - 1 is the last state id assigned. */
 	maxState = redFsm->nextStateId - 1;
 
+	for ( CondSpaceList::Iter csi = cgd->condSpaceList; csi.lte(); csi++ ) {
+		if ( csi->condSpaceId > maxCondSpaceId )
+			maxCondSpaceId = csi->condSpaceId;
+	}
+
 	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Maximum cond length. */
+		if ( st->stateCondList.length() > maxCondLen )
+			maxCondLen = st->stateCondList.length();
+
 		/* Maximum single length. */
 		if ( st->outSingle.length() > maxSingleLen )
 			maxSingleLen = st->outSingle.length();
@@ -163,21 +186,35 @@ void FsmCodeGen::setValueLimits()
 
 		/* The key offset index offset for the state after last is not used, skip it.. */
 		if ( ! st.last() ) {
+			maxCondOffset += st->stateCondList.length();
 			maxKeyOffset += st->outSingle.length() + st->outRange.length()*2;
 			maxIndexOffset += st->outSingle.length() + st->outRange.length() + 1;
 		}
 
+		/* Max cond span. */
+		if ( st->condList != 0 ) {
+			unsigned long long span = keyOps->span( st->condLowKey, st->condHighKey );
+			if ( span > maxCondSpan )
+				maxCondSpan = span;
+		}
+
 		/* Max key span. */
 		if ( st->transList != 0 ) {
-			unsigned long long span = redFsm->keyOps->span( st->lowKey, st->highKey );
+			unsigned long long span = keyOps->span( st->lowKey, st->highKey );
 			if ( span > maxSpan )
 				maxSpan = span;
 		}
 
+		/* Max cond index offset. */
+		if ( ! st.last() ) {
+			if ( st->condList != 0 )
+				maxCondIndexOffset += keyOps->span( st->condLowKey, st->condHighKey );
+		}
+
 		/* Max flat index offset. */
 		if ( ! st.last() ) {
 			if ( st->transList != 0 )
-				maxFlatIndexOffset += redFsm->keyOps->span( st->lowKey, st->highKey );
+				maxFlatIndexOffset += keyOps->span( st->lowKey, st->highKey );
 			maxFlatIndexOffset += 1;
 		}
 	}
@@ -206,7 +243,9 @@ void FsmCodeGen::analyzeAction( Action *act, InlineList *inlineList )
 	for ( InlineList::Iter item = *inlineList; item.lte(); item++ ) {
 		/* Only consider actions that are referenced. */
 		if ( act->numRefs() > 0 ) {
-			if ( item->type == InlineItem::Call )
+			if ( item->type == InlineItem::Goto || item->type == InlineItem::GotoExpr )
+				bAnyActionGotos = true;
+			else if ( item->type == InlineItem::Call || item->type == InlineItem::CallExpr )
 				bAnyActionCalls = true;
 			else if ( item->type == InlineItem::Ret )
 				bAnyActionRets = true;
@@ -309,7 +348,7 @@ void FsmCodeGen::analyzeMachine()
 				st->defTrans->action->anyCurStateRef() )
 			st->bAnyRegCurStateRef = true;
 		
-		if ( st->stateConds.length() > 0 )
+		if ( st->stateCondList.length() > 0 )
 			bAnyConditions = true;
 	}
 
@@ -323,24 +362,37 @@ void FsmCodeGen::analyzeMachine()
 	calcIndexSize();
 }
 
-int FsmCodeGen::arrayTypeSize( unsigned long maxVal )
+unsigned int FsmCodeGen::arrayTypeSize( unsigned long maxVal )
 {
-	if ( maxVal <= RL_UCHAR_MAX )
-		return 1;
-	else if ( maxVal <= RL_USHORT_MAX )
-		return 2;
-	else 
-		return 4;
+	long long maxValLL = (long long) maxVal;
+	HostType *arrayType = keyOps->typeSubsumes( maxValLL );
+	assert( arrayType != 0 );
+	return arrayType->size;
 }
 
+string FsmCodeGen::ARRAY_TYPE( unsigned long maxVal )
+{
+	long long maxValLL = (long long) maxVal;
+	HostType *arrayType = keyOps->typeSubsumes( maxValLL );
+	assert( arrayType != 0 );
+
+	string ret = arrayType->data1;
+	if ( arrayType->data2 != 0 ) {
+		ret += " ";
+		ret += arrayType->data2;
+	}
+	return ret;
+}
+
+
 /* Write out the fsm name. */
-std::string FsmCodeGen::FSM_NAME()
+string FsmCodeGen::FSM_NAME()
 {
 	return fsmName;
 }
 
 /* Emit the offset of the start state as a decimal integer. */
-std::string FsmCodeGen::START_STATE_ID()
+string FsmCodeGen::START_STATE_ID()
 {
 	ostringstream ret;
 	ret << redFsm->startState->id;
@@ -408,7 +460,7 @@ string FsmCodeGen::GET_WIDE_KEY()
 
 string FsmCodeGen::GET_WIDE_KEY( RedStateAp *state )
 {
-	if ( state->stateConds.length() > 0 )
+	if ( state->stateCondList.length() > 0 )
 		return "_widec";
 	else
 		return GET_KEY();
@@ -442,13 +494,13 @@ string FsmCodeGen::TABS( int level )
 
 /* Write out a key from the fsm code gen. Depends on wether or not the key is
  * signed. */
-string FsmCodeGen::KEY( long key )
+string FsmCodeGen::KEY( Key key )
 {
 	ostringstream ret;
-	if ( cgd->isAlphSigned() )
-		ret << key;
+	if ( keyOps->isSigned || !hostLang->explicitUnsigned )
+		ret << key.getVal();
 	else
-		ret << (unsigned long) key << 'u';
+		ret << (unsigned long) key.getVal() << 'u';
 	return ret.str();
 }
 
@@ -653,7 +705,7 @@ void FsmCodeGen::CONDITION( ostream &ret, Action *condition )
 	INLINE_LIST( ret, condition->inlineList, 0, false );
 }
 
-std::string FsmCodeGen::ERROR_STATE()
+string FsmCodeGen::ERROR_STATE()
 {
 	ostringstream ret;
 	if ( redFsm->errState != 0 )
@@ -698,61 +750,50 @@ string FsmCodeGen::DATA_PREFIX()
 	return "";
 }
 
-void FsmCodeGen::COND_TRANSLATE( RedStateAp *state, bool genDefault )
+/* Emit the alphabet data type. */
+string FsmCodeGen::ALPH_TYPE()
 {
-	out << "		switch ( " << GET_KEY() << " ) {\n";
-	for ( CondVect::Iter sc = state->stateConds; sc.lte(); sc++ ) {
-		Condition *condition = *sc;
-		out << 
-			TABS(2) << "case " << condition->key << ":\n" <<
-			TABS(3) << "_widec = " << condition->baseKey << ";\n";
-
-		for ( CondSet::Iter csi = condition->condSet; csi.lte(); csi++ ) {
-			out << TABS(3) << "if ( ";
-			CONDITION( out, *csi );
-			out << " ) _widec += " << ( 1 << csi.pos() ) << ";\n";
-		}
-
-		out <<
-			TABS(3) << "break;\n";
+	string ret = keyOps->alphType->data1;
+	if ( keyOps->alphType->data2 != 0 ) {
+		ret += " ";
+		ret += + keyOps->alphType->data2;
 	}
-
-	if ( genDefault ) {
-		out << 
-			TABS(2) << "default:\n" <<
-			TABS(3) << "_widec = " << GET_KEY() << ";\n" <<
-			TABS(3) << "break;\n";
-	}
-		
-	out <<
-		"		}\n";
+	return ret;
 }
 
-void FsmCodeGen::COND_TRANSLATE()
+/* Emit the alphabet data type. */
+string FsmCodeGen::WIDE_ALPH_TYPE()
 {
-	out << 
-		"	_widec = " << GET_KEY() << ";\n"
-		"	switch( " << CS() << " ) {\n";
+	string ret;
+	if ( maxKey <= keyOps->maxKey )
+		ret = ALPH_TYPE();
+	else {
+		long long maxKeyVal = maxKey.getLongLong();
+		HostType *wideType = keyOps->typeSubsumes( keyOps->isSigned, maxKeyVal );
+		assert( wideType != 0 );
 
-	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
-		if ( st->stateConds.length() > 0 ) {
-			out << "	case " << st->id << ":\n";
-			COND_TRANSLATE( st, false );
-			out << "		break;\n";
+		ret = wideType->data1;
+		if ( wideType->data2 != 0 ) {
+			ret += " ";
+			ret += wideType->data2;
 		}
 	}
-
-	out <<
-		"	}\n";
+	return ret;
 }
 
+
 /*
  * Language specific, but style independent code generators functions.
  */
 
-std::ostream &CCodeGen::OPEN_ARRAY( std::string type, string name )
+string CCodeGen::PTR_CONST()
 {
-	out << "static " << type << " " << name << "[] = {\n";
+	return "const ";
+}
+
+std::ostream &CCodeGen::OPEN_ARRAY( string type, string name )
+{
+	out << "static const " << type << " " << name << "[] = {\n";
 	return out;
 }
 
@@ -761,6 +802,12 @@ std::ostream &CCodeGen::CLOSE_ARRAY()
 	return out << "};\n";
 }
 
+std::ostream &CCodeGen::STATIC_VAR( string type, string name )
+{
+	out << "static const " << type << " " << name;
+	return out;
+}
+
 string CCodeGen::UINT( )
 {
 	return "unsigned int";
@@ -776,24 +823,12 @@ string CCodeGen::CAST( string type )
 	return "(" + type + ")";
 }
 
-std::string CCodeGen::ARRAY_TYPE( unsigned long maxVal )
-{
-	if ( maxVal <= RL_UCHAR_MAX )
-		return "unsigned char";
-	else if ( maxVal <= RL_USHORT_MAX )
-		return "unsigned short";
-	else if ( maxVal <= RL_UINT_MAX )
-		return "unsigned int";
-	else 
-		return "unsigned long";
-}
-
-std::string CCodeGen::NULL_ITEM()
+string CCodeGen::NULL_ITEM()
 {
 	return "0";
 }
 
-std::string CCodeGen::POINTER()
+string CCodeGen::POINTER()
 {
 	return " *";
 }
@@ -803,96 +838,29 @@ std::ostream &CCodeGen::SWITCH_DEFAULT()
 	return out;
 }
 
-/* Emit the alphabet data type. */
-std::string CCodeGen::ALPH_TYPE()
-{
-	switch ( cgd->alphType ) {
-	case AT_Char:
-		return "char";
-	case AT_UnsignedChar:
-		return "unsigned char";
-	case AT_Short:
-		return "short";
-	case AT_UnsignedShort:
-		return "unsigned short";
-	case AT_Int:
-		return "int";
-	case AT_UnsignedInt:
-		return "unsigned int";
-	case AT_Long:
-		return "long";
-	case AT_UnsignedLong:
-		return "unsigned long";
-	}
-	return "";
-}
-
-/* Emit the alphabet data type. */
-std::string CCodeGen::WIDE_ALPH_TYPE()
-{
-	if ( redFsm->keyOps->gt( maxKey, redFsm->keyOps->highKey ) ) {
-		if ( cgd->isAlphSigned() ) {
-			if ( redFsm->keyOps->lteq( maxKey, RL_CHAR_MAX ) )
-				return "char";
-			else if ( redFsm->keyOps->lteq( maxKey, RL_SHORT_MAX ) )
-				return "short";
-			else if ( redFsm->keyOps->lteq( maxKey, RL_INT_MAX ) )
-				return "int";
-			else 
-				return "long";
-		}
-		else {
-			if ( redFsm->keyOps->lteq( maxKey, RL_UCHAR_MAX ) )
-				return "unsigned char";
-			else if ( redFsm->keyOps->lteq( maxKey, RL_USHORT_MAX ) )
-				return "unsigned short";
-			else if ( redFsm->keyOps->lteq( maxKey, RL_UINT_MAX ) )
-				return "unsigned int";
-			else 
-				return "unsigned long";
-		}
-	}
-	else {
-		switch ( cgd->alphType ) {
-		case AT_Char:
-			return "char";
-		case AT_UnsignedChar:
-			return "unsigned char";
-		case AT_Short:
-			return "short";
-		case AT_UnsignedShort:
-			return "unsigned short";
-		case AT_Int:
-			return "int";
-		case AT_UnsignedInt:
-			return "unsigned int";
-		case AT_Long:
-			return "long";
-		case AT_UnsignedLong:
-			return "unsigned long";
-		}
-	}
-	return "";
-}
-
 /*
  * D Specific
  */
 
-std::string DCodeGen::NULL_ITEM()
+string DCodeGen::NULL_ITEM()
 {
 	return "null";
 }
 
-std::string DCodeGen::POINTER()
+string DCodeGen::POINTER()
 {
 	// multiple items seperated by commas can also be pointer types.
 	return "* ";
 }
 
-std::ostream &DCodeGen::OPEN_ARRAY( std::string type, string name )
+string DCodeGen::PTR_CONST()
 {
-	out << "static " << type << "[] " << name << " = [\n";
+	return "";
+}
+
+std::ostream &DCodeGen::OPEN_ARRAY( string type, string name )
+{
+	out << "static const " << type << "[] " << name << " = [\n";
 	return out;
 }
 
@@ -901,53 +869,10 @@ std::ostream &DCodeGen::CLOSE_ARRAY()
 	return out << "];\n";
 }
 
-std::string DCodeGen::ARRAY_TYPE( unsigned long maxVal )
+std::ostream &DCodeGen::STATIC_VAR( string type, string name )
 {
-	if ( maxVal <= RL_UCHAR_MAX )
-		return "ubyte";
-	else if ( maxVal <= RL_USHORT_MAX )
-		return "ushort";
-	else 
-		return "uint";
-}
-
-
-/* Emit the alphabet data type for the D language. */
-std::string DCodeGen::ALPH_TYPE()
-{
-	switch ( cgd->alphType ) {
-	case AT_Char:
-	case AT_UnsignedChar:
-		return "char";
-	case AT_Short:
-	case AT_UnsignedShort:
-		return "wchar";
-	case AT_Int:
-	case AT_UnsignedInt:
-	case AT_Long:
-	case AT_UnsignedLong:
-		return "dchar";
-	}
-	return "";
-}
-
-/* Emit the alphabet data type for the D language. */
-std::string DCodeGen::WIDE_ALPH_TYPE()
-{
-	switch ( cgd->alphType ) {
-	case AT_Char:
-	case AT_UnsignedChar:
-		return "char";
-	case AT_Short:
-	case AT_UnsignedShort:
-		return "wchar";
-	case AT_Int:
-	case AT_UnsignedInt:
-	case AT_Long:
-	case AT_UnsignedLong:
-		return "dchar";
-	}
-	return "";
+	out << "static const " << type << " " << name;
+	return out;
 }
 
 string DCodeGen::ARR_OFF( string ptr, string offset )
@@ -970,3 +895,83 @@ std::ostream &DCodeGen::SWITCH_DEFAULT()
 	out << "		default: break;\n";
 	return out;
 }
+
+/* 
+ * Java Specific
+ */
+
+string JavaCodeGen::PTR_CONST()
+{
+	/* Not used in Java code. */
+	assert( false );
+	return "final";
+}
+
+std::ostream &JavaCodeGen::OPEN_ARRAY( string type, string name )
+{
+	out << "static final " << type << "[] " << name << " = {\n";
+	return out;
+}
+
+std::ostream &JavaCodeGen::CLOSE_ARRAY()
+{
+	return out << "};\n";
+}
+
+std::ostream &JavaCodeGen::STATIC_VAR( string type, string name )
+{
+	out << "static final " << type << " " << name;
+	return out;
+}
+
+string JavaCodeGen::UINT( )
+{
+	/* Not used. */
+	assert( false );
+	return "long";
+}
+
+string JavaCodeGen::ARR_OFF( string ptr, string offset )
+{
+	return ptr + " + " + offset;
+}
+
+string JavaCodeGen::CAST( string type )
+{
+	return "(" + type + ")";
+}
+
+string JavaCodeGen::NULL_ITEM()
+{
+	/* Not used. */
+	assert( false );
+	return "0";
+}
+
+string JavaCodeGen::POINTER()
+{
+	/* Not used. */
+	assert( false );
+	return " *";
+}
+
+std::ostream &JavaCodeGen::SWITCH_DEFAULT()
+{
+	return out;
+}
+
+string JavaCodeGen::GET_KEY()
+{
+	ostringstream ret;
+	if ( cgd->getKeyExpr != 0 ) { 
+		/* Emit the user supplied method of retrieving the key. */
+		ret << "(";
+		INLINE_LIST( ret, cgd->getKeyExpr, 0, false );
+		ret << ")";
+	}
+	else {
+		/* Expression for retrieving the key, use simple dereference. */
+		ret << "data[" << P() << "]";
+	}
+	return ret.str();
+}
