@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -27,6 +27,7 @@
 #include <iostream>
 #include <string>
 #include <stdio.h>
+#include "common.h"
 
 using std::string;
 using std::ostream;
@@ -71,34 +72,30 @@ public:
 	void analyzeAction( Action *act, InlineList *inlineList );
 	void analyzeActionList( RedAction *redAct, InlineList *inlineList );
 	void analyzeMachine();
-	
+
 protected:
 	friend struct CodeGenData;
 
 	string FSM_NAME();
 	string START_STATE_ID();
 	ostream &ACTIONS_ARRAY();
-	string GET_KEY();
 	string GET_WIDE_KEY();
 	string GET_WIDE_KEY( RedStateAp *state );
 	string TABS( int level );
-	string KEY( long key );
+	string KEY( Key key );
 	string LDIR_PATH( char *path );
 	void ACTION( ostream &ret, Action *action, int targState, bool inFinish );
 	void CONDITION( ostream &ret, Action *condition );
-
-	void COND_TRANSLATE( RedStateAp *state, bool genDefault );
-	void COND_TRANSLATE();
-
+	string ALPH_TYPE();
+	string WIDE_ALPH_TYPE();
+	string ARRAY_TYPE( unsigned long maxVal );
 
 	virtual string ARR_OFF( string ptr, string offset ) = 0;
 	virtual string CAST( string type ) = 0;
-	virtual string UINT( ) = 0;
-	virtual string ARRAY_TYPE( unsigned long maxVal ) = 0;
-	virtual string ALPH_TYPE() = 0;
-	virtual string WIDE_ALPH_TYPE() = 0;
+	virtual string UINT() = 0;
 	virtual string NULL_ITEM() = 0;
 	virtual string POINTER() = 0;
+	virtual string GET_KEY();
 	virtual ostream &SWITCH_DEFAULT() = 0;
 
 	string P() { return "p"; }
@@ -113,19 +110,25 @@ protected:
 	string ACT() { return ACCESS() + "act"; }
 
 	string DATA_PREFIX();
-	string K() { return "_" + DATA_PREFIX() + "k"; }
-	string I() { return "_" + DATA_PREFIX() + "i"; }
-	string KO() { return "_" + DATA_PREFIX() + "ko"; }
-	string IO() { return "_" + DATA_PREFIX() + "io"; }
-	string SL() { return "_" + DATA_PREFIX() + "sl"; }
-	string RL() { return "_" + DATA_PREFIX() + "rl"; }
-	string A() { return "_" + DATA_PREFIX() + "a"; }
-	string TA() { return "_" + DATA_PREFIX() + "ta"; }
-	string TT() { return "_" + DATA_PREFIX() + "tt"; }
-	string TSA() { return "_" + DATA_PREFIX() + "tsa"; }
-	string FSA() { return "_" + DATA_PREFIX() + "fsa"; }
-	string EA() { return "_" + DATA_PREFIX() + "ea"; }
-	string SP() { return "_" + DATA_PREFIX() + "sp"; }
+	string PM() { return "_" + DATA_PREFIX() + "partition_map"; }
+	string C() { return "_" + DATA_PREFIX() + "cond_spaces"; }
+	string CK() { return "_" + DATA_PREFIX() + "cond_keys"; }
+	string K() { return "_" + DATA_PREFIX() + "trans_keys"; }
+	string I() { return "_" + DATA_PREFIX() + "indicies"; }
+	string CO() { return "_" + DATA_PREFIX() + "cond_offsets"; }
+	string KO() { return "_" + DATA_PREFIX() + "key_offsets"; }
+	string IO() { return "_" + DATA_PREFIX() + "index_offsets"; }
+	string CL() { return "_" + DATA_PREFIX() + "cond_lengths"; }
+	string SL() { return "_" + DATA_PREFIX() + "single_lengths"; }
+	string RL() { return "_" + DATA_PREFIX() + "range_lengths"; }
+	string A() { return "_" + DATA_PREFIX() + "actions"; }
+	string TA() { return "_" + DATA_PREFIX() + "trans_actions_wi"; }
+	string TT() { return "_" + DATA_PREFIX() + "trans_targs_wi"; }
+	string TSA() { return "_" + DATA_PREFIX() + "to_state_actions"; }
+	string FSA() { return "_" + DATA_PREFIX() + "from_state_actions"; }
+	string EA() { return "_" + DATA_PREFIX() + "eof_actions"; }
+	string SP() { return "_" + DATA_PREFIX() + "key_spans"; }
+	string CSP() { return "_" + DATA_PREFIX() + "cond_key_spans"; }
 	string START() { return DATA_PREFIX() + "start"; }
 	string ERROR() { return DATA_PREFIX() + "error"; }
 	string FIRST_FINAL() { return DATA_PREFIX() + "first_final"; }
@@ -157,16 +160,19 @@ protected:
 	string ERROR_STATE();
 	string FIRST_FINAL_STATE();
 
+	virtual string PTR_CONST() = 0;
 	virtual ostream &OPEN_ARRAY( string type, string name ) = 0;
 	virtual ostream &CLOSE_ARRAY() = 0;
+	virtual ostream &STATIC_VAR( string type, string name ) = 0;
 
-	int arrayTypeSize( unsigned long maxVal );
+	unsigned int arrayTypeSize( unsigned long maxVal );
 
 	bool anyActions();
 	bool anyToStateActions()        { return bAnyToStateActions; }
 	bool anyFromStateActions()      { return bAnyFromStateActions; }
 	bool anyRegActions()            { return bAnyRegActions; }
 	bool anyEofActions()            { return bAnyEofActions; }
+	bool anyActionGotos()           { return bAnyActionGotos; }
 	bool anyActionCalls()           { return bAnyActionCalls; }
 	bool anyActionRets()            { return bAnyActionRets; }
 	bool anyRegActionRets()         { return bAnyRegActionRets; }
@@ -193,6 +199,9 @@ protected:
 	CodeGenData *cgd;
 	RedFsmAp *redFsm;
 
+	bool outLabelUsed;
+	bool againLabelUsed;
+
 protected:
 	ostream &out;
 
@@ -200,6 +209,7 @@ protected:
 	bool bAnyFromStateActions;
 	bool bAnyRegActions;
 	bool bAnyEofActions;
+	bool bAnyActionGotos;
 	bool bAnyActionCalls;
 	bool bAnyActionRets;
 	bool bAnyRegActionRets;
@@ -220,8 +230,14 @@ protected:
 	int maxActionLoc;
 	int maxActArrItem;
 	unsigned long long maxSpan;
+	unsigned long long maxCondSpan;
 	int maxFlatIndexOffset;
-	long maxKey;
+	Key maxKey;
+	int maxCondOffset;
+	int maxCondLen;
+	int maxCondSpaceId;
+	int maxCondIndexOffset;
+	int maxCond;
 
 	bool useIndicies;
 };
@@ -229,33 +245,47 @@ protected:
 class CCodeGen : virtual public FsmCodeGen
 {
 public:
-	virtual string ARRAY_TYPE( unsigned long maxVal );
-	virtual string ALPH_TYPE();
-	virtual string WIDE_ALPH_TYPE();
 	virtual string NULL_ITEM();
 	virtual string POINTER();
 	virtual ostream &SWITCH_DEFAULT();
 	virtual ostream &OPEN_ARRAY( string type, string name );
 	virtual ostream &CLOSE_ARRAY();
+	virtual ostream &STATIC_VAR( string type, string name );
 	virtual string ARR_OFF( string ptr, string offset );
 	virtual string CAST( string type );
-	virtual string UINT( );
+	virtual string UINT();
+	virtual string PTR_CONST();
 };
 
 class DCodeGen : virtual public FsmCodeGen
 {
 public:
-	virtual string ARRAY_TYPE( unsigned long maxVal );
-	virtual string ALPH_TYPE();
-	virtual string WIDE_ALPH_TYPE();
 	virtual string NULL_ITEM();
 	virtual string POINTER();
 	virtual ostream &SWITCH_DEFAULT();
 	virtual ostream &OPEN_ARRAY( string type, string name );
 	virtual ostream &CLOSE_ARRAY();
+	virtual ostream &STATIC_VAR( string type, string name );
 	virtual string ARR_OFF( string ptr, string offset );
 	virtual string CAST( string type );
-	virtual string UINT( );
+	virtual string UINT();
+	virtual string PTR_CONST();
+};
+
+class JavaCodeGen : virtual public FsmCodeGen
+{
+public:
+	virtual string NULL_ITEM();
+	virtual string POINTER();
+	virtual ostream &SWITCH_DEFAULT();
+	virtual ostream &OPEN_ARRAY( string type, string name );
+	virtual ostream &CLOSE_ARRAY();
+	virtual ostream &STATIC_VAR( string type, string name );
+	virtual string ARR_OFF( string ptr, string offset );
+	virtual string CAST( string type );
+	virtual string UINT();
+	virtual string PTR_CONST();
+	virtual string GET_KEY();
 };
 
 #endif /* _FSMCODEGEN_H */
