@@ -1,5 +1,5 @@
 /*
- *  Copyright 2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2005-2006 Adrian Thurston <thurston@cs.queensu.ca>
  */
 
 /*  This file is part of Ragel.
@@ -30,6 +30,8 @@
 #include "gotocodegen.h"
 #include "fgotocodegen.h"
 #include "ipgotocodegen.h"
+#include "splitcodegen.h"
+#include "javacodegen.h"
 
 #include <iostream>
 
@@ -40,8 +42,7 @@ CodeGenData *cgd = 0;
 
 void CodeGenData::createMachine()
 {
-	setKeyOps();
-	redFsm = new RedFsmAp( &keyOps );
+	redFsm = new RedFsmAp();
 }
 
 void CodeGenData::initActionList( unsigned long length )
@@ -90,8 +91,8 @@ void CodeGenData::initTransList( int snum, unsigned long length )
 	 * nothing. */
 }
 
-void CodeGenData::newTrans( int snum, int tnum, long lowKey, 
-		long highKey, long targ, long action )
+void CodeGenData::newTrans( int snum, int tnum, Key lowKey, 
+		Key highKey, long targ, long action )
 {
 	/* Get the current state and range. */
 	RedStateAp *curState = allStates + snum;
@@ -109,25 +110,25 @@ void CodeGenData::newTrans( int snum, int tnum, long lowKey,
 		 * the error transitions. */
 		if ( destRange.length() == 0 ) {
 			/* Range is currently empty. */
-			if ( keyOps.lt( keyOps.lowKey, lowKey ) ) {
+			if ( keyOps->minKey < lowKey ) {
 				/* The first range doesn't start at the low end. */
-				long fillHighKey = lowKey;
-				keyOps.dec( fillHighKey );
+				Key fillHighKey = lowKey;
+				fillHighKey.decrement();
 
 				/* Create the filler with the state's error transition. */
-				RedTransEl newTel( keyOps.lowKey, fillHighKey, redFsm->getErrorTrans() );
+				RedTransEl newTel( keyOps->minKey, fillHighKey, redFsm->getErrorTrans() );
 				destRange.append( newTel );
 			}
 		}
 		else {
 			/* The range list is not empty, get the the last range. */
 			RedTransEl *last = &destRange[destRange.length()-1];
-			long nextKey = last->highKey;
-			keyOps.inc( nextKey );
-			if ( keyOps.lt( nextKey, lowKey ) ) {
+			Key nextKey = last->highKey;
+			nextKey.increment();
+			if ( nextKey < lowKey ) {
 				/* There is a gap to fill. Make the high key. */
-				long fillHighKey = lowKey;
-				keyOps.dec( fillHighKey );
+				Key fillHighKey = lowKey;
+				fillHighKey.decrement();
 
 				/* Create the filler with the state's error transtion. */
 				RedTransEl newTel( nextKey, fillHighKey, redFsm->getErrorTrans() );
@@ -152,19 +153,19 @@ void CodeGenData::finishTransList( int snum )
 		if ( destRange.length() == 0 ) {
 			/* Fill with the whole alphabet. */
 			/* Add the range on the lower and upper bound. */
-			RedTransEl newTel( keyOps.lowKey, keyOps.highKey, redFsm->getErrorTrans() );
+			RedTransEl newTel( keyOps->minKey, keyOps->maxKey, redFsm->getErrorTrans() );
 			destRange.append( newTel );
 		}
 		else {
 			/* Get the last and check for a gap on the end. */
 			RedTransEl *last = &destRange[destRange.length()-1];
-			if ( keyOps.lt( last->highKey, keyOps.highKey ) ) {
+			if ( last->highKey < keyOps->maxKey ) {
 				/* Make the high key. */
-				long fillLowKey = last->highKey;
-				keyOps.inc( fillLowKey );
+				Key fillLowKey = last->highKey;
+				fillLowKey.increment();
 
 				/* Create the new range with the error trans and append it. */
-				RedTransEl newTel( fillLowKey, keyOps.highKey, redFsm->getErrorTrans() );
+				RedTransEl newTel( fillLowKey, keyOps->maxKey, redFsm->getErrorTrans() );
 				destRange.append( newTel );
 			}
 		}
@@ -224,65 +225,71 @@ void CodeGenData::finishMachine()
 	/* Note that even if we want a complete graph we do not give the error
 	 * state a default transition. All machines break out of the processing
 	 * loop when in the error state. */
+
+	if ( codeStyle == GenGoto || codeStyle == GenFGoto || codeStyle == GenIpGoto ) {
+		for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+			for ( StateCondList::Iter sci = st->stateCondList; sci.lte(); sci++ )
+				st->stateCondVect.append( sci );
+		}
+	}
 }
 
 
 bool CodeGenData::setAlphType( char *data )
 {
-	bool valid = true;
-	if ( strcmp( "signed_char", data ) == 0 )
-		alphType = AT_Char;
-	else if ( strcmp( "unsigned_char", data ) == 0 )
-		alphType = AT_UnsignedChar;
-	else if ( strcmp( "signed_short", data ) == 0 )
-		alphType = AT_Short;
-	else if ( strcmp( "unsigned_short", data ) == 0 )
-		alphType = AT_UnsignedShort;
-	else if ( strcmp( "signed_int", data ) == 0 )
-		alphType = AT_Int;
-	else if ( strcmp( "unsigned_int", data ) == 0 )
-		alphType = AT_UnsignedInt;
-	else if ( strcmp( "signed_long", data ) == 0 )
-		alphType = AT_Long;
-	else if ( strcmp( "unsigned_long", data ) == 0 )
-		alphType = AT_UnsignedLong;
-	else
-		valid = false;
-	return valid;
+	/* FIXME: This should validate the alphabet type selection. */
+	HostType *alphType = hostLang->hostTypes + atoi(data);
+	thisKeyOps.setAlphType( alphType );
+	return true;
 }
 
-void CodeGenData::addStateCond( int snum, ulong cond )
+void CodeGenData::initCondSpaceList( ulong length )
 {
-	RedStateAp *curState = allStates + snum;
-	curState->stateConds.append( allConditions + cond );
-}
-
-
-void CodeGenData::initCondList( ulong length )
-{
-	allConditions = new Condition[length];
+	allCondSpaces = new CondSpace[length];
 	for ( ulong c = 0; c < length; c++ )
-		conditionList.append( allConditions + c );
+		condSpaceList.append( allCondSpaces + c );
 }
 
-void CodeGenData::newCondition( int cnum, long key, long baseKey )
+void CodeGenData::newCondSpace( int cnum, int condSpaceId, Key baseKey )
 {
-	Condition *cond = allConditions + cnum;
-	cond->key = key;
+	CondSpace *cond = allCondSpaces + cnum;
+	cond->condSpaceId = condSpaceId;
 	cond->baseKey = baseKey;
 }
 
-void CodeGenData::conditionItem( int cnum, long actionId )
+void CodeGenData::condSpaceItem( int cnum, long condActionId )
 {
-	Condition *cond = allConditions + cnum;
-	cond->condSet.append( allActions + actionId );
+	CondSpace *cond = allCondSpaces + cnum;
+	cond->condSet.append( allActions + condActionId );
 }
 
+void CodeGenData::initStateCondList( int snum, ulong length )
+{
+	/* Could preallocate these, as we could with transitions. */
+}
+
+void CodeGenData::addStateCond( int snum, Key lowKey, Key highKey, long condNum )
+{
+	RedStateAp *curState = allStates + snum;
+
+	/* Create the new state condition. */
+	StateCond *stateCond = new StateCond;
+	stateCond->lowKey = lowKey;
+	stateCond->highKey = highKey;
+
+	/* Assign it a cond space. */
+	CondSpace *condSpace = allCondSpaces + condNum;
+	stateCond->condSpace = condSpace;
+
+	curState->stateCondList.append( stateCond );
+}
+
+
 /* Generate the codegen depending on the command line options given. */
 void CodeGenData::makeCodeGen()
 {
-	switch ( outputFormat ) {
-	case OutCCode:
+	switch ( hostLangType ) {
+	case CCode:
 		switch ( codeStyle ) {
 		case GenTables:
 			codeGen = new CTabCodeGen;
@@ -305,9 +312,13 @@ void CodeGenData::makeCodeGen()
 		case GenIpGoto:
 			codeGen = new CIpGotoCodeGen;
 			break;
+		case GenSplit:
+			codeGen = new CSplitCodeGen;
+			break;
 		}
 		break;
-	case OutDCode:
+
+	case DCode:
 		switch ( codeStyle ) {
 		case GenTables:
 			codeGen = new DTabCodeGen;
@@ -330,88 +341,61 @@ void CodeGenData::makeCodeGen()
 		case GenIpGoto:
 			codeGen = new DIpGotoCodeGen;
 			break;
+		case GenSplit:
+			codeGen = new DSplitCodeGen;
+			break;
 		}
 		break;
 
-	/* Called codegen when generating something else. */
-	default:
-		assert( false );
+	case JavaCode:
+		switch ( codeStyle ) {
+		case GenTables:
+			codeGen = new JavaTabCodeGen;
+			break;
+		default:
+			assert(false);
+			break;
+		}
+		break;
 	}
 
 	codeGen->fsmName = fsmName;
 	codeGen->cgd = this;
 }
 
-bool CodeGenData::isAlphSigned()
+CondSpace *CodeGenData::findCondSpace( Key lowKey, Key highKey )
 {
-	return alphType == AT_Char || alphType == AT_Short || alphType == AT_Int;
-}
+	for ( CondSpaceList::Iter cs = condSpaceList; cs.lte(); cs++ ) {
+		Key csHighKey = cs->baseKey;
+		csHighKey += keyOps->alphSize() * (1 << cs->condSet.length());
 
-void CodeGenData::setKeyOps( )
-{
-	/* Use default ranges for alphabet type. */
-	switch ( alphType ) {
-	case AT_Char:
-		lowKey = RL_CHAR_MIN;
-		highKey = RL_CHAR_MAX;
-		break;
-	case AT_UnsignedChar:
-		lowKey = RL_UCHAR_MIN;
-		highKey = RL_UCHAR_MAX;
-		break;
-	case AT_Short:
-		lowKey = RL_SHORT_MIN;
-		highKey = RL_SHORT_MAX;
-		break;
-	case AT_UnsignedShort:
-		lowKey = RL_USHORT_MIN;
-		highKey = RL_USHORT_MAX;
-		break;
-	case AT_Int:
-		lowKey = RL_INT_MIN;
-		highKey = RL_INT_MAX;
-		break;
-	case AT_UnsignedInt:
-		lowKey = RL_UINT_MIN;
-		highKey = RL_UINT_MAX;
-		break;
-	case AT_Long:
-		lowKey = RL_LONG_MIN;
-		highKey = RL_LONG_MAX;
-		break;
-	case AT_UnsignedLong:
-		lowKey = RL_ULONG_MIN;
-		highKey = RL_ULONG_MAX;
-		break;
+		if ( lowKey >= cs->baseKey && highKey <= csHighKey )
+			return cs;
 	}
-
-	/* Signedness and bounds. */
-	keyOps.isAlphSigned = isAlphSigned();
-	keyOps.lowKey = lowKey;
-	keyOps.highKey = highKey;
+	return 0;
 }
 
-Condition *CodeGenData::findCondition( long key )
+Condition *CodeGenData::findCondition( Key key )
 {
 	for ( ConditionList::Iter cond = conditionList; cond.lte(); cond++ ) {
-		long upperKey = cond->baseKey + (1 << cond->condSet.length());
+		Key upperKey = cond->baseKey + (1 << cond->condSet.length());
 		if ( cond->baseKey <= key && key <= upperKey )
 			return cond;
 	}
 	return 0;
 }
 
-long CodeGenData::findMaxKey()
+Key CodeGenData::findMaxKey()
 {
-	long maxKey = redFsm->keyOps->highKey;
+	Key maxKey = keyOps->maxKey;
 	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
 		assert( st->outSingle.length() == 0 );
 		assert( st->defTrans == 0 );
 
 		long rangeLen = st->outRange.length();
 		if ( rangeLen > 0 ) {
-			long highKey = st->outRange[rangeLen-1].highKey;
-			if ( keyOps.gt( highKey, maxKey ) )
+			Key highKey = st->outRange[rangeLen-1].highKey;
+			if ( highKey > maxKey )
 				maxKey = highKey;
 		}
 	}
@@ -428,14 +412,16 @@ void CodeGenData::prepareMachine()
 	
 	/* Do this before distributing transitions out to singles and defaults
 	 * makes life easier. */
-	long maxKey = findMaxKey();
+	Key maxKey = findMaxKey();
 
 	redFsm->assignActionLocs();
 
 	/* Order the states. */
 	redFsm->depthFirstOrdering();
 
-	if ( codeStyle == GenGoto || codeStyle == GenFGoto || codeStyle == GenIpGoto ) {
+	if ( codeStyle == GenGoto || codeStyle == GenFGoto || 
+			codeStyle == GenIpGoto || codeStyle == GenSplit )
+	{
 		/* For goto driven machines we can keep the original depth
 		 * first ordering because it's ok if the state ids are not
 		 * sequential. Split the the ids by final state status. */
@@ -467,6 +453,12 @@ void CodeGenData::prepareMachine()
 	/* If any errors have occured in the input file then don't write anything. */
 	if ( gblErrorCount > 0 )
 		return;
+	
+	if ( codeStyle == GenSplit )
+		redFsm->partitionFsm( numSplitPartitions );
+
+	if ( codeStyle == GenIpGoto || codeStyle == GenSplit )
+		redFsm->setInTrans();
 
 	/* Make a code generator that will output the header/code. */
 	if ( codeGen == 0 )
@@ -490,7 +482,7 @@ void CodeGenData::generateGraphviz()
 	redFsm->chooseDefaultSpan();
 
 	/* Make the generator. */
-	GraphvizDotGen dotGen( fsmName, this, redFsm, std::cout );
+	GraphvizDotGen dotGen( fsmName, this, redFsm, *outStream );
 
 	/* Write out with it. */
 	dotGen.writeDotFile();
@@ -518,19 +510,22 @@ void CodeGenData::generateCode()
 		genLineDirective( *outStream );
 	}
 	
-	if ( writeData )
-		codeGen->writeOutData();
-
-	if ( writeInit )
-		codeGen->writeOutInit();
 
 	if ( writeExec ) {
 		/* Must set labels immediately before writing because we may depend
 		 * on the noend write option. */
 		codeGen->setLabelsNeeded();
-		codeGen->writeOutExec();
 	}
 
+	if ( writeData )
+		codeGen->writeOutData();
+	
+	if ( writeInit )
+		codeGen->writeOutInit();
+
+	if ( writeExec )
+		codeGen->writeOutExec();
+
 	if ( writeEOF )
 		codeGen->writeOutEOF();
 }
@@ -538,7 +533,7 @@ void CodeGenData::generateCode()
 void CodeGenData::generate()
 {
 	if ( redFsm != 0 ) {
-		if ( outputFormat == OutCCode || outputFormat == OutDCode )
+		if ( outputFormat == OutCode )
 			generateCode();
 		else if ( outputFormat == OutGraphvizDot && !graphvizDone ) {
 			graphvizDone = true;
@@ -549,15 +544,17 @@ void CodeGenData::generate()
 
 void lineDirective( ostream &out, char *fileName, int line )
 {
-	/* Write the preprocessor line info for to the input file. */
-	out << "#line " << line  << " \"";
-	for ( char *pc = fileName; *pc != 0; pc++ ) {
-		if ( *pc == '\\' )
-			out << "\\\\";
-		else
-			out << *pc;
+	if ( hostLangType != JavaCode ) {
+		/* Write the preprocessor line info for to the input file. */
+		out << "#line " << line  << " \"";
+		for ( char *pc = fileName; *pc != 0; pc++ ) {
+			if ( *pc == '\\' )
+				out << "\\\\";
+			else
+				out << *pc;
+		}
+		out << "\"\n";
 	}
-	out << "\"\n";
 }
 
 void genLineDirective( ostream &out )
