@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -117,7 +117,8 @@ void GotoCodeGen::emitSingleSwitch( RedStateAp *state )
 
 	if ( numSingles == 1 ) {
 		/* If there is a single single key then write it out as an if. */
-		out << "\tif ( " << GET_WIDE_KEY(state) << " == " << KEY(data[0].lowKey) << " )\n\t\t"; 
+		out << "\tif ( " << GET_WIDE_KEY(state) << " == " << 
+				KEY(data[0].lowKey) << " )\n\t\t"; 
 
 		/* Virtual function for writing the target of the transition. */
 		TRANS_GOTO(data[0].value, 0) << "\n";
@@ -151,21 +152,24 @@ void GotoCodeGen::emitRangeBSearch( RedStateAp *state, int level, int low, int h
 	bool anyHigher = mid < high;
 
 	/* Determine if the keys at mid are the limits of the alphabet. */
-	bool limitLow = cgd->redFsm->keyOps->eq( data[mid].lowKey, cgd->lowKey );
-	bool limitHigh = cgd->redFsm->keyOps->eq( data[mid].highKey, cgd->highKey );
+	bool limitLow = data[mid].lowKey == keyOps->minKey;
+	bool limitHigh = data[mid].highKey == keyOps->maxKey;
 
 	if ( anyLower && anyHigher ) {
 		/* Can go lower and higher than mid. */
-		out << TABS(level) << "if ( " << GET_WIDE_KEY(state) << " < " << KEY(data[mid].lowKey) << " ) {\n";
+		out << TABS(level) << "if ( " << GET_WIDE_KEY(state) << " < " << 
+				KEY(data[mid].lowKey) << " ) {\n";
 		emitRangeBSearch( state, level+1, low, mid-1 );
-		out << TABS(level) << "} else if ( " << GET_WIDE_KEY(state) << " > " << KEY(data[mid].highKey) << " ) {\n";
+		out << TABS(level) << "} else if ( " << GET_WIDE_KEY(state) << " > " << 
+				KEY(data[mid].highKey) << " ) {\n";
 		emitRangeBSearch( state, level+1, mid+1, high );
 		out << TABS(level) << "} else\n";
 		TRANS_GOTO(data[mid].value, level+1) << "\n";
 	}
 	else if ( anyLower && !anyHigher ) {
 		/* Can go lower than mid but not higher. */
-		out << TABS(level) << "if ( " << GET_WIDE_KEY(state) << " < " << KEY(data[mid].lowKey) << " ) {\n";
+		out << TABS(level) << "if ( " << GET_WIDE_KEY(state) << " < " << 
+				KEY(data[mid].lowKey) << " ) {\n";
 		emitRangeBSearch( state, level+1, low, mid-1 );
 
 		/* if the higher is the highest in the alphabet then there is no
@@ -175,13 +179,15 @@ void GotoCodeGen::emitRangeBSearch( RedStateAp *state, int level, int low, int h
 			TRANS_GOTO(data[mid].value, level+1) << "\n";
 		}
 		else {
-			out << TABS(level) << "} else if ( " << GET_WIDE_KEY(state) << " <= " << KEY(data[mid].highKey) << " )\n";
+			out << TABS(level) << "} else if ( " << GET_WIDE_KEY(state) << " <= " << 
+					KEY(data[mid].highKey) << " )\n";
 			TRANS_GOTO(data[mid].value, level+1) << "\n";
 		}
 	}
 	else if ( !anyLower && anyHigher ) {
 		/* Can go higher than mid but not lower. */
-		out << TABS(level) << "if ( " << GET_WIDE_KEY(state) << " > " << KEY(data[mid].highKey) << " ) {\n";
+		out << TABS(level) << "if ( " << GET_WIDE_KEY(state) << " > " << 
+				KEY(data[mid].highKey) << " ) {\n";
 		emitRangeBSearch( state, level+1, mid+1, high );
 
 		/* If the lower end is the lowest in the alphabet then there is no
@@ -191,7 +197,8 @@ void GotoCodeGen::emitRangeBSearch( RedStateAp *state, int level, int low, int h
 			TRANS_GOTO(data[mid].value, level+1) << "\n";
 		}
 		else {
-			out << TABS(level) << "} else if ( " << GET_WIDE_KEY(state) << " >= " << KEY(data[mid].lowKey) << " )\n";
+			out << TABS(level) << "} else if ( " << GET_WIDE_KEY(state) << " >= " << 
+					KEY(data[mid].lowKey) << " )\n";
 			TRANS_GOTO(data[mid].value, level+1) << "\n";
 		}
 	}
@@ -199,16 +206,19 @@ void GotoCodeGen::emitRangeBSearch( RedStateAp *state, int level, int low, int h
 		/* Cannot go higher or lower than mid. It's mid or bust. What
 		 * tests to do depends on limits of alphabet. */
 		if ( !limitLow && !limitHigh ) {
-			out << TABS(level) << "if ( " << KEY(data[mid].lowKey) << " <= " << GET_WIDE_KEY(state) << 
-				" && " << GET_WIDE_KEY(state) << " <= " << KEY(data[mid].highKey) << " )\n";
+			out << TABS(level) << "if ( " << KEY(data[mid].lowKey) << " <= " << 
+					GET_WIDE_KEY(state) << " && " << GET_WIDE_KEY(state) << " <= " << 
+					KEY(data[mid].highKey) << " )\n";
 			TRANS_GOTO(data[mid].value, level+1) << "\n";
 		}
 		else if ( limitLow && !limitHigh ) {
-			out << TABS(level) << "if ( " << GET_WIDE_KEY(state) << " <= " << KEY(data[mid].highKey) << " )\n";
+			out << TABS(level) << "if ( " << GET_WIDE_KEY(state) << " <= " << 
+					KEY(data[mid].highKey) << " )\n";
 			TRANS_GOTO(data[mid].value, level+1) << "\n";
 		}
 		else if ( !limitLow && limitHigh ) {
-			out << TABS(level) << "if ( " << KEY(data[mid].lowKey) << " <= " << GET_WIDE_KEY(state) << " )\n";
+			out << TABS(level) << "if ( " << KEY(data[mid].lowKey) << " <= " << 
+					GET_WIDE_KEY(state) << " )\n";
 			TRANS_GOTO(data[mid].value, level+1) << "\n";
 		}
 		else {
@@ -221,11 +231,121 @@ void GotoCodeGen::emitRangeBSearch( RedStateAp *state, int level, int low, int h
 void GotoCodeGen::STATE_GOTO_ERROR()
 {
 	/* Label the state and bail immediately. */
+	outLabelUsed = true;
 	RedStateAp *state = redFsm->errState;
 	out << "case " << state->id << ":\n";
 	out << "	goto _out;\n";
 }
 
+void GotoCodeGen::COND_TRANSLATE( StateCond *stateCond, int level )
+{
+	CondSpace *condSpace = stateCond->condSpace;
+	out << TABS(level) << "_widec = " << KEY(condSpace->baseKey) << 
+			" + (" << GET_KEY() << " - " << KEY(keyOps->minKey) << ");\n";
+
+	for ( CondSet::Iter csi = condSpace->condSet; csi.lte(); csi++ ) {
+		out << TABS(level) << "if ( ";
+		CONDITION( out, *csi );
+		Size condValOffset = ((1 << csi.pos()) * keyOps->alphSize());
+		out << " ) _widec += " << condValOffset << ";\n";
+	}
+}
+
+void GotoCodeGen::emitCondBSearch( RedStateAp *state, int level, int low, int high )
+{
+	/* Get the mid position, staying on the lower end of the range. */
+	int mid = (low + high) >> 1;
+	StateCond **data = state->stateCondVect.data;
+
+	/* Determine if we need to look higher or lower. */
+	bool anyLower = mid > low;
+	bool anyHigher = mid < high;
+
+	/* Determine if the keys at mid are the limits of the alphabet. */
+	bool limitLow = data[mid]->lowKey == keyOps->minKey;
+	bool limitHigh = data[mid]->highKey == keyOps->maxKey;
+
+	if ( anyLower && anyHigher ) {
+		/* Can go lower and higher than mid. */
+		out << TABS(level) << "if ( " << GET_KEY() << " < " << 
+				KEY(data[mid]->lowKey) << " ) {\n";
+		emitCondBSearch( state, level+1, low, mid-1 );
+		out << TABS(level) << "} else if ( " << GET_KEY() << " > " << 
+				KEY(data[mid]->highKey) << " ) {\n";
+		emitCondBSearch( state, level+1, mid+1, high );
+		out << TABS(level) << "} else {\n";
+		COND_TRANSLATE(data[mid], level+1);
+		out << TABS(level) << "}\n";
+	}
+	else if ( anyLower && !anyHigher ) {
+		/* Can go lower than mid but not higher. */
+		out << TABS(level) << "if ( " << GET_KEY() << " < " << 
+				KEY(data[mid]->lowKey) << " ) {\n";
+		emitCondBSearch( state, level+1, low, mid-1 );
+
+		/* if the higher is the highest in the alphabet then there is no
+		 * sense testing it. */
+		if ( limitHigh ) {
+			out << TABS(level) << "} else {\n";
+			COND_TRANSLATE(data[mid], level+1);
+			out << TABS(level) << "}\n";
+		}
+		else {
+			out << TABS(level) << "} else if ( " << GET_KEY() << " <= " << 
+					KEY(data[mid]->highKey) << " ) {\n";
+			COND_TRANSLATE(data[mid], level+1);
+			out << TABS(level) << "}\n";
+		}
+	}
+	else if ( !anyLower && anyHigher ) {
+		/* Can go higher than mid but not lower. */
+		out << TABS(level) << "if ( " << GET_KEY() << " > " << 
+				KEY(data[mid]->highKey) << " ) {\n";
+		emitCondBSearch( state, level+1, mid+1, high );
+
+		/* If the lower end is the lowest in the alphabet then there is no
+		 * sense testing it. */
+		if ( limitLow ) {
+			out << TABS(level) << "} else {\n";
+			COND_TRANSLATE(data[mid], level+1);
+			out << TABS(level) << "}\n";
+		}
+		else {
+			out << TABS(level) << "} else if ( " << GET_KEY() << " >= " << 
+					KEY(data[mid]->lowKey) << " ) {\n";
+			COND_TRANSLATE(data[mid], level+1);
+			out << TABS(level) << "}\n";
+		}
+	}
+	else {
+		/* Cannot go higher or lower than mid. It's mid or bust. What
+		 * tests to do depends on limits of alphabet. */
+		if ( !limitLow && !limitHigh ) {
+			out << TABS(level) << "if ( " << KEY(data[mid]->lowKey) << " <= " << 
+					GET_KEY() << " && " << GET_KEY() << " <= " << 
+					KEY(data[mid]->highKey) << " ) {\n";
+			COND_TRANSLATE(data[mid], level+1);
+			out << TABS(level) << "}\n";
+		}
+		else if ( limitLow && !limitHigh ) {
+			out << TABS(level) << "if ( " << GET_KEY() << " <= " << 
+					KEY(data[mid]->highKey) << " ) {\n";
+			COND_TRANSLATE(data[mid], level+1);
+			out << TABS(level) << "}\n";
+		}
+		else if ( !limitLow && limitHigh ) {
+			out << TABS(level) << "if ( " << KEY(data[mid]->lowKey) << " <= " << 
+					GET_KEY() << " )\n {";
+			COND_TRANSLATE(data[mid], level+1);
+			out << TABS(level) << "}\n";
+		}
+		else {
+			/* Both high and low are at the limit. No tests to do. */
+			COND_TRANSLATE(data[mid], level);
+		}
+	}
+}
+
 std::ostream &GotoCodeGen::STATE_GOTOS()
 {
 	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
@@ -235,8 +355,10 @@ std::ostream &GotoCodeGen::STATE_GOTOS()
 			/* Writing code above state gotos. */
 			GOTO_HEADER( st );
 
-			if ( st->stateConds.length() > 0 )
-				COND_TRANSLATE( st, true );
+			if ( st->stateCondVect.length() > 0 ) {
+				out << "	_widec = " << GET_KEY() << ";\n";
+				emitCondBSearch( st, 1, 0, st->stateCondVect.length() - 1 );
+			}
 
 			/* Try singles. */
 			if ( st->outSingle.length() > 0 )
@@ -472,24 +594,25 @@ void GotoCodeGen::RET( ostream &ret, bool inFinish )
 
 void GotoCodeGen::BREAK( ostream &ret, int targState )
 {
+	outLabelUsed = true;
 	ret << "goto _out;";
 }
 
 void GotoCodeGen::writeOutData()
 {
 	out <<
-		"static int " << START() << " = " << START_STATE_ID() << ";\n"
+		"static const int " << START() << " = " << START_STATE_ID() << ";\n"
 		"\n";
 
 	if ( cgd->writeFirstFinal ) {
 		out <<
-			"static int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
+			"static const int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
 			"\n";
 	}
 
 	if ( cgd->writeErr ) {
 		out <<
-			"static int " << ERROR() << " = " << ERROR_STATE() << ";\n"
+			"static const int " << ERROR() << " = " << ERROR_STATE() << ";\n"
 			"\n";
 	}
 
@@ -524,14 +647,16 @@ void GotoCodeGen::writeOutData()
 
 void GotoCodeGen::writeOutExec()
 {
+	outLabelUsed = false;
+
 	out << "	{\n";
 
 	if ( anyRegCurStateRef() )
 		out << "	int _ps = 0;\n";
 
-	if ( anyActions() ) {
+	if ( anyToStateActions() || anyRegActions() || anyFromStateActions() ) {
 		out << 
-			"	" << ARRAY_TYPE(maxActArrItem) << " *_acts;\n"
+			"	" << PTR_CONST() << ARRAY_TYPE(maxActArrItem) << POINTER() << "_acts;\n"
 			"	" << UINT() << " _nacts;\n";
 	}
 
@@ -540,39 +665,15 @@ void GotoCodeGen::writeOutExec()
 
 	out << "\n";
 
-	out << 
-		"	" << P() << " -= 1;\n"
-		"	goto _resume;\n"
-		"\n";
-
-	/* Head of the loop. */
-	out << "_again:\n";
-
-	if ( anyToStateActions() ) {
-		out <<
-			"	_acts = " << ARR_OFF( A(), TSA() + "[" + CS() + "]" ) << ";\n"
-			"	_nacts = " << CAST(UINT()) << " *_acts++;\n"
-			"	while ( _nacts-- > 0 ) {\n"
-			"		switch ( *_acts++ ) {\n";
-			TO_STATE_ACTION_SWITCH();
-			SWITCH_DEFAULT() <<
-			"		}\n"
-			"	}\n"
-			"\n";
+	if ( cgd->hasEnd ) {
+		outLabelUsed = true;
+		out << 
+			"	if ( " << P() << " == " << PE() << " )\n"
+			"		goto _out;\n";
 	}
 
 	out << "_resume:\n";
 
-	if ( cgd->hasEnd ) {
-		out << 
-			"	if ( ++" << P() << " == " << PE() << " )\n"
-			"		goto _out;\n";
-	}
-	else {
-		out << 
-			"	" << P() << " += 1;\n";
-	}
-
 	if ( anyFromStateActions() ) {
 		out <<
 			"	_acts = " << ARR_OFF( A(), FSA() + "[" + CS() + "]" ) << ";\n"
@@ -598,7 +699,33 @@ void GotoCodeGen::writeOutExec()
 	if ( anyRegActions() )
 		EXEC_FUNCS() << "\n";
 
-	if ( cgd->hasEnd || anyRegBreak() || redFsm->errState != 0 )
+	out << "_again:\n";
+
+	if ( anyToStateActions() ) {
+		out <<
+			"	_acts = " << ARR_OFF( A(), TSA() + "[" + CS() + "]" ) << ";\n"
+			"	_nacts = " << CAST(UINT()) << " *_acts++;\n"
+			"	while ( _nacts-- > 0 ) {\n"
+			"		switch ( *_acts++ ) {\n";
+			TO_STATE_ACTION_SWITCH();
+			SWITCH_DEFAULT() <<
+			"		}\n"
+			"	}\n"
+			"\n";
+	}
+
+	if ( cgd->hasEnd ) {
+		out << 
+			"	if ( ++" << P() << " != " << PE() << " )\n"
+			"		goto _resume;\n";
+	}
+	else {
+		out << 
+			"	" << P() << " += 1;\n"
+			"	goto _resume;\n";
+	}
+
+	if ( outLabelUsed )
 		out << "	_out: {}\n";
 
 	out << "	}\n";
@@ -609,7 +736,7 @@ void GotoCodeGen::writeOutEOF()
 	if ( anyEofActions() ) {
 		out << 
 			"	{\n"
-			"	" << ARRAY_TYPE(maxActArrItem) << " *_acts = " << 
+			"	" << PTR_CONST() << ARRAY_TYPE(maxActArrItem) << POINTER() << "_acts = " << 
 					ARR_OFF( A(), EA() + "[" + CS() + "]" ) << ";\n"
 			"	" << UINT() << " _nacts = " << CAST(UINT()) << " *_acts++;\n"
 			"	while ( _nacts-- > 0 ) {\n"
