@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -32,6 +32,7 @@ struct CodeGenData;
 struct NameInst;
 struct RedTransAp;
 struct RedStateAp;
+struct StateCond;
 
 /*
  * Goto driven fsm.
@@ -67,6 +68,10 @@ public:
 	std::ostream &FROM_STATE_ACTIONS();
 	std::ostream &EOF_ACTIONS();
 
+	void COND_TRANSLATE( StateCond *stateCond, int level );
+	void emitCondBSearch( RedStateAp *state, int level, int low, int high );
+	void STATE_CONDS( RedStateAp *state, bool genDefault ); 
+
 	virtual std::ostream &TRANS_GOTO( RedTransAp *trans, int level );
 
 	void emitSingleSwitch( RedStateAp *state );
