@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2004 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  */
 
 /*  This file is part of Ragel.
@@ -37,12 +37,29 @@ GraphvizDotGen::GraphvizDotGen( char *fsmName, CodeGenData *cgd,
 {
 }
 
-std::ostream &GraphvizDotGen::KEY( long key )
+std::ostream &GraphvizDotGen::KEY( Key key )
 {
-	if ( cgd->isAlphSigned() )
-		out << key;
-	else
-		out << (unsigned long) key;
+	if ( printPrintables && key.isPrintable() ) {
+		// Output values as characters, ensuring we escape the quote (") character
+		char cVal = (char) key.getVal();
+		out << "'";
+		switch ( cVal ) {
+			case '"': case '\\':
+				out << "\\" << cVal;
+				break;
+			default:	
+				out << cVal;
+				break;
+		}
+		out << "'";
+	}
+	else {
+		if ( keyOps->isSigned )
+			out << key.getVal();
+		else
+			out << (unsigned long) key.getVal();
+	}
+
 	return out;
 }
 
@@ -65,10 +82,7 @@ std::ostream &GraphvizDotGen::TRANS_ACTION( RedStateAp *fromState, RedTransAp *t
 	for ( int a = 0; a < n; a++ ) {
 		for ( ActionTable::Iter actIt = actions[a]->key.first(); actIt.lte(); actIt++ ) {
 			Action *action = actIt->value;
-			if ( action->name != 0 )
-				out << action->name;
-			else
-				out << action->loc.line << ":" << action->loc.col;
+			out << action->nameOrLoc();
 			if ( a < n-1 || !actIt.last() )
 				out << ", ";
 		}
@@ -92,31 +106,37 @@ std::ostream &GraphvizDotGen::ACTION( RedAction *action )
 	return out;
 }
 
-std::ostream &GraphvizDotGen::ONCHAR( long lowKey, long highKey )
+std::ostream &GraphvizDotGen::ONCHAR( Key lowKey, Key highKey )
 {
-	KeyOps *keyOps = redFsm->keyOps;
-	if ( keyOps->gt( lowKey, keyOps->highKey ) ) {
-		for ( long key = lowKey; keyOps->lteq( key, highKey ); keyOps->inc( key ) ) {
-			Condition *cond = cgd->findCondition( key );
-			KEY( cond->key );
-			out << "(";
+	if ( lowKey > keyOps->maxKey ) {
+		CondSpace *condSpace = cgd->findCondSpace( lowKey, highKey );
+		Key values = ( lowKey - condSpace->baseKey ) / keyOps->alphSize();
 
-			long values = key - cond->baseKey;
-			for ( CondSet::Iter csi = cond->condSet; csi.lte(); csi++ ) {
-				bool set = values & (1 << csi.pos());
-				if ( !set )
-					out << "!";
-				out << (*csi)->name;
-				if ( !csi.last() )
-					out << ", ";
-			}
-			out << ")";
+		lowKey = keyOps->minKey + 
+			(lowKey - condSpace->baseKey - keyOps->alphSize() * values.getVal());
+		highKey = keyOps->minKey + 
+			(highKey - condSpace->baseKey - keyOps->alphSize() * values.getVal());
+		KEY( lowKey );
+		if ( lowKey != highKey ) {
+			out << "..";
+			KEY( highKey );
 		}
+		out << "(";
+
+		for ( CondSet::Iter csi = condSpace->condSet; csi.lte(); csi++ ) {
+			bool set = values & (1 << csi.pos());
+			if ( !set )
+				out << "!";
+			out << (*csi)->nameOrLoc();
+			if ( !csi.last() )
+				out << ", ";
+		}
+		out << ")";
 	}
 	else {
 		/* Output the key. Possibly a range. */
 		KEY( lowKey );
-		if ( ! redFsm->keyOps->eq( highKey, lowKey ) ) {
+		if ( highKey != lowKey ) {
 			out << "..";
 			KEY( highKey );
 		}
