@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -84,14 +84,14 @@ void IpGotoCodeGen::BREAK( ostream &ret, int targState )
 	ret << "goto _out" << targState << ";";
 }
 
-bool IpGotoCodeGen::TRANS_ACTIONS( RedStateAp *state )
+bool IpGotoCodeGen::IN_TRANS_ACTIONS( RedStateAp *state )
 {
 	bool anyWritten = false;
 
-	/* Emit any transitions that have actions and that go to 
-	 * this state. */
-	for ( TransApSet::Iter trans = redFsm->transSet; trans.lte(); trans ++ ) {
-		if ( trans->targ == state && trans->action != 0 ) {
+	/* Emit any transitions that have actions and that go to this state. */
+	for ( int it = 0; it < state->numInTrans; it++ ) {
+		RedTransAp *trans = state->inTrans[it];
+		if ( trans->action != 0 && trans->labelNeeded ) {
 			/* Remember that we wrote an action so we know to write the
 			 * line directive for going back to the output. */
 			anyWritten = true;
@@ -124,7 +124,7 @@ bool IpGotoCodeGen::TRANS_ACTIONS( RedStateAp *state )
  * state. */
 void IpGotoCodeGen::GOTO_HEADER( RedStateAp *state )
 {
-	bool anyWritten = TRANS_ACTIONS( state );
+	bool anyWritten = IN_TRANS_ACTIONS( state );
 
 	if ( state->labelNeeded ) 
 		out << "st" << state->id << ":\n";
@@ -172,7 +172,7 @@ void IpGotoCodeGen::STATE_GOTO_ERROR()
 	/* In the error state we need to emit some stuff that usually goes into
 	 * the header. */
 	RedStateAp *state = redFsm->errState;
-	bool anyWritten = TRANS_ACTIONS( state );
+	bool anyWritten = IN_TRANS_ACTIONS( state );
 
 	/* No case label needed since we don't switch on the error state. */
 	if ( anyWritten )
@@ -204,6 +204,7 @@ std::ostream &IpGotoCodeGen::EXIT_STATES()
 {
 	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
 		if ( st->outNeeded ) {
+			outLabelUsed = true;
 			out << "	_out" << st->id << ": " << CS() << " = " << 
 					st->id << "; goto _out; \n";
 		}
@@ -225,18 +226,24 @@ std::ostream &IpGotoCodeGen::FINISH_CASES()
 	bool anyWritten = false;
 
 	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
-		/* Cased needed if there are out actions. */
 		if ( st->eofAction != 0 ) {
-			out << "\tcase " << st->id << ": ";
+			if ( st->eofAction->eofRefs == 0 )
+				st->eofAction->eofRefs = new IntSet;
+			st->eofAction->eofRefs->insert( st->id );
+		}
+	}
+
+	for ( ActionTableMap::Iter act = redFsm->actionMap; act.lte(); act++ ) {
+		if ( act->eofRefs != 0 ) {
+			for ( IntSet::Iter pst = *act->eofRefs; pst.lte(); pst++ )
+				out << "	case " << *pst << ": \n";
 
-			/* If there are out functions, write them. */
 			/* Remember that we wrote a trans so we know to write the
 			 * line directive for going back to the output. */
 			anyWritten = true;
 
-			out << "\n";
 			/* Write each action in the eof action list. */
-			for ( ActionTable::Iter item = st->eofAction->key; item.lte(); item++ )
+			for ( ActionTable::Iter item = act->key; item.lte(); item++ )
 				ACTION( out, item->value, STATE_ERR_STATE, true );
 			out << "\tbreak;\n";
 		}
@@ -320,24 +327,26 @@ void IpGotoCodeGen::setLabelsNeeded()
 void IpGotoCodeGen::writeOutData()
 {
 	out <<
-		"static int " << START() << " = " << START_STATE_ID() << ";\n"
+		"static const int " << START() << " = " << START_STATE_ID() << ";\n"
 		"\n";
 
 	if ( cgd->writeFirstFinal ) {
 		out <<
-			"static int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
+			"static const int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
 			"\n";
 	}
 
 	if ( cgd->writeErr ) {
 		out <<
-			"static int " << ERROR() << " = " << ERROR_STATE() << ";\n"
+			"static const int " << ERROR() << " = " << ERROR_STATE() << ";\n"
 			"\n";
 	}
 }
 
 void IpGotoCodeGen::writeOutExec()
 {
+	outLabelUsed = false;
+
 	out << "	{\n";
 
 	if ( anyRegCurStateRef() )
@@ -346,8 +355,12 @@ void IpGotoCodeGen::writeOutExec()
 	if ( anyConditions() )
 		out << "	" << WIDE_ALPH_TYPE() << " _widec;\n";
 
-	out << 
-		"	" << P() << " -= 1;\n";
+	if ( cgd->hasEnd ) {
+		outLabelUsed = true;
+		out << 
+			"	if ( " << P() << " == " << PE() << " )\n"
+			"		goto _out;\n";
+	}
 
 	if ( useAgainLabel() ) {
 		out << 
@@ -358,18 +371,20 @@ void IpGotoCodeGen::writeOutExec()
 			AGAIN_CASES() <<
 			"	default: break;\n"
 			"	}\n"
-			"\n"
-			"_resume:\n";
-	}
+			"\n";
 
-	if ( cgd->hasEnd ) {
-		out << 
-			"	if ( ++" << P() << " == " << PE() << " )\n"
-			"		goto _out;\n";
-	}
-	else {
-		out << 
-			"	" << P() << " += 1;\n";
+		if ( cgd->hasEnd ) {
+			outLabelUsed = true;
+			out << 
+				"	if ( ++" << P() << " == " << PE() << " )\n"
+				"		goto _out;\n";
+		}
+		else {
+			out << 
+				"	" << P() << " += 1;\n";
+		}
+
+		out << "_resume:\n";
 	}
 
 	out << 
@@ -378,8 +393,12 @@ void IpGotoCodeGen::writeOutExec()
 		SWITCH_DEFAULT() <<
 		"	}\n";
 		EXIT_STATES() << 
-		"\n"
-		"	_out: {}\n"
+		"\n";
+
+	if ( outLabelUsed ) 
+		out << "	_out: {}\n";
+
+	out <<
 		"	}\n";
 }
 
