@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2004 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  */
 
 /*  This file is part of Ragel.
@@ -23,6 +23,8 @@
 #define _REDFSM_H
 
 #include <assert.h>
+#include <string.h>
+#include <string>
 #include "common.h"
 #include "vector.h"
 #include "dlist.h"
@@ -42,6 +44,8 @@
 #define STATE_ERR_STATE   0
 #define FUNC_NO_FUNC      0
 
+using std::string;
+
 struct RedStateAp;
 struct InlineList;
 struct Action;
@@ -104,6 +108,8 @@ struct Action
 	InlineList *inlineList;
 	int actionId;
 
+	string nameOrLoc();
+
 	/* Number of references in the final machine. */
 	int numRefs() 
 		{ return numTransRefs + numToStateRefs + numFromStateRefs + numEofRefs; }
@@ -116,7 +122,6 @@ struct Action
 
 /* Forwards. */
 struct RedStateAp;
-struct TransAp;
 struct StateAp;
 
 /* Transistion Action Element. */
@@ -152,6 +157,9 @@ struct CmpActionTableEl
 /* Compare for ActionTable. */
 typedef CmpSTable< ActionTableEl, CmpActionTableEl > CmpActionTable;
 
+/* Set of states. */
+typedef BstSet<RedStateAp*> RedStateSet;
+typedef BstSet<int> IntSet;
 
 /* Reduced action. */
 struct RedAction
@@ -161,6 +169,7 @@ struct RedAction
 	RedAction( )
 	:	
 		key(), 
+		eofRefs(0),
 		numTransRefs(0),
 		numToStateRefs(0),
 		numFromStateRefs(0),
@@ -176,6 +185,7 @@ struct RedAction
 	ActionTable key;
 	int actListId;
 	int location;
+	IntSet *eofRefs;
 
 	/* Number of references in the final machine. */
 	bool numRefs() 
@@ -201,11 +211,13 @@ struct RedTransAp
 	public AvlTreeEl<RedTransAp>
 {
 	RedTransAp( RedStateAp *targ, RedAction *action, int id )
-		: targ(targ), action(action), id(id) { }
+		: targ(targ), action(action), id(id), labelNeeded(true) { }
 
 	RedStateAp *targ;
 	RedAction *action;
 	int id;
+	bool partitionBoundary;
+	bool labelNeeded;
 };
 
 /* Compare of transitions for the final reduction of transitions. Comparison
@@ -234,10 +246,10 @@ typedef AvlBasic<RedTransAp, CmpRedTransAp> TransApSet;
 struct RedTransEl
 {
 	/* Constructors. */
-	RedTransEl( long lowKey, long highKey, RedTransAp *value ) 
+	RedTransEl( Key lowKey, Key highKey, RedTransAp *value ) 
 		: lowKey(lowKey), highKey(highKey), value(value) { }
 
-	long lowKey, highKey;
+	Key lowKey, highKey;
 	RedTransAp *value;
 };
 
@@ -267,25 +279,43 @@ typedef MergeSort<RedSpanMapEl, CmpRedSpanMapEl> RedSpanMapSort;
 /* Set of entry ids that go into this state. */
 typedef Vector<int> EntryIdVect;
 typedef Vector<char*> EntryNameVect;
-typedef BstSet<RedStateAp*> RedStateSet;
 
-typedef BstSet< Action*, CmpOrd<Action*> > CondSet;
+typedef Vector< Action* > CondSet;
 
 struct Condition
 {
 	Condition( )
 		: key(0), baseKey(0) {}
 
-	long key;
-	long baseKey;
+	Key key;
+	Key baseKey;
 	CondSet condSet;
 
 	Condition *next, *prev;
 };
-
 typedef DList<Condition> ConditionList;
 
-typedef Vector<Condition*> CondVect;
+struct CondSpace
+{
+	Key baseKey;
+	CondSet condSet;
+	int condSpaceId;
+
+	CondSpace *next, *prev;
+};
+typedef DList<CondSpace> CondSpaceList;
+
+struct StateCond
+{
+	Key lowKey;
+	Key highKey;
+
+	CondSpace *condSpace;
+
+	StateCond *prev, *next;
+};
+typedef DList<StateCond> StateCondList;
+typedef Vector<StateCond*> StateCondVect;
 
 /* Reduced state. */
 struct RedStateAp
@@ -293,6 +323,7 @@ struct RedStateAp
 	RedStateAp()
 	: 
 		defTrans(0), 
+		condList(0),
 		transList(0), 
 		isFinal(false), 
 		labelNeeded(false), 
@@ -302,7 +333,10 @@ struct RedStateAp
 		fromStateAction(0), 
 		eofAction(0), 
 		id(0), 
-		bAnyRegCurStateRef(false)
+		bAnyRegCurStateRef(false),
+		partitionBoundary(false),
+		inTrans(0),
+		numInTrans(0)
 	{ }
 
 	/* Transitions out. */
@@ -310,8 +344,12 @@ struct RedStateAp
 	RedTransList outRange;
 	RedTransAp *defTrans;
 
+	/* For flat conditions. */
+	Key condLowKey, condHighKey;
+	CondSpace **condList;
+
 	/* For flat keys. */
-	long lowKey, highKey;
+	Key lowKey, highKey;
 	RedTransAp **transList;
 
 	/* The list of states that transitions from this state go to. */
@@ -325,13 +363,20 @@ struct RedStateAp
 	RedAction *fromStateAction;
 	RedAction *eofAction;
 	int id;
-	CondVect stateConds;
+	StateCondList stateCondList;
+	StateCondVect stateCondVect;
 
 	/* Pointers for the list of states. */
 	RedStateAp *prev, *next;
 
 	bool anyRegCurStateRef() { return bAnyRegCurStateRef; }
 	bool bAnyRegCurStateRef;
+
+	int partition;
+	bool partitionBoundary;
+
+	RedTransAp **inTrans;
+	int numInTrans;
 };
 
 /* List of states. */
@@ -343,11 +388,10 @@ typedef BstSet< RedTransAp*, CmpOrd<RedTransAp*> > RedTransSet;
 /* Next version of the fsm machine. */
 struct RedFsmAp
 {
-	RedFsmAp( KeyOps *keyOps );
+	RedFsmAp();
 
 	bool wantComplete;
 	bool forcedErrorState;
-	KeyOps *keyOps;
 
 	int nextActionId;
 	int nextTransId;
@@ -365,6 +409,7 @@ struct RedFsmAp
 	RedTransAp *errActionTrans;
 	RedStateAp *firstFinState;
 	int numFinStates;
+	int nParts;
 
 	/* Is is it possible to extend a range by bumping ranges that span only
 	 * one character to the singles array. */
@@ -419,6 +464,10 @@ struct RedFsmAp
 	bool alphabetCovered( RedTransList &outRange );
 
 	RedTransAp *allocateTrans( RedStateAp *targState, RedAction *actionTable );
+
+	void partitionFsm( int nParts );
+
+	void setInTrans();
 };
 
 
