@@ -1,5 +1,5 @@
 /*
- *  Copyright 2001-2003, 2005 Adrian Thurston <thurston@cs.queensu.ca>
+ *  Copyright 2001-2006 Adrian Thurston <thurston@cs.queensu.ca>
  *            2004 Eric Ocean <eric.ocean@ampede.com>
  *            2005 Alan West <alan@alanz.com>
  */
@@ -161,6 +161,26 @@ std::ostream &TabCodeGen::ACTION_SWITCH()
 	return out;
 }
 
+std::ostream &TabCodeGen::COND_OFFSETS()
+{
+	out << "\t";
+	int totalStateNum = 0, curKeyOffset = 0;
+	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Write the key offset. */
+		out << curKeyOffset;
+		if ( !st.last() ) {
+			out << ", ";
+			if ( ++totalStateNum % IALL == 0 )
+				out << "\n\t";
+		}
+
+		/* Move the key offset ahead. */
+		curKeyOffset += st->stateCondList.length();
+	}
+	out << "\n";
+	return out;
+}
+
 std::ostream &TabCodeGen::KEY_OFFSETS()
 {
 	out << "\t";
@@ -204,6 +224,24 @@ std::ostream &TabCodeGen::INDEX_OFFSETS()
 	return out;
 }
 
+std::ostream &TabCodeGen::COND_LENS()
+{
+	out << "\t";
+	int totalStateNum = 0;
+	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Write singles length. */
+		out << st->stateCondList.length();
+		if ( !st.last() ) {
+			out << ", ";
+			if ( ++totalStateNum % IALL == 0 )
+				out << "\n\t";
+		}
+	}
+	out << "\n";
+	return out;
+}
+
+
 std::ostream &TabCodeGen::SINGLE_LENS()
 {
 	out << "\t";
@@ -289,6 +327,51 @@ std::ostream &TabCodeGen::EOF_ACTIONS()
 	return out;
 }
 
+std::ostream &TabCodeGen::COND_KEYS()
+{
+	out << '\t';
+	int totalTrans = 0;
+	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Loop the state's transitions. */
+		for ( StateCondList::Iter sc = st->stateCondList; sc.lte(); sc++ ) {
+			/* Lower key. */
+			out << KEY( sc->lowKey ) << ", ";
+			if ( ++totalTrans % IALL == 0 )
+				out << "\n\t";
+
+			/* Upper key. */
+			out << KEY( sc->highKey ) << ", ";
+			if ( ++totalTrans % IALL == 0 )
+				out << "\n\t";
+		}
+	}
+
+	/* Output one last number so we don't have to figure out when the last
+	 * entry is and avoid writing a comma. */
+	out << 0 << "\n";
+	return out;
+}
+
+std::ostream &TabCodeGen::COND_SPACES()
+{
+	out << '\t';
+	int totalTrans = 0;
+	for ( RedStateList::Iter st = redFsm->stateList; st.lte(); st++ ) {
+		/* Loop the state's transitions. */
+		for ( StateCondList::Iter sc = st->stateCondList; sc.lte(); sc++ ) {
+			/* Cond Space id. */
+			out << sc->condSpace->condSpaceId << ", ";
+			if ( ++totalTrans % IALL == 0 )
+				out << "\n\t";
+		}
+	}
+
+	/* Output one last number so we don't have to figure out when the last
+	 * entry is and avoid writing a comma. */
+	out << 0 << "\n";
+	return out;
+}
+
 std::ostream &TabCodeGen::KEYS()
 {
 	out << '\t';
@@ -484,11 +567,11 @@ void TabCodeGen::LOCATE_TRANS()
 		"	_keys = " << ARR_OFF( K(), KO() + "[" + CS() + "]" ) << ";\n"
 		"	_trans = " << IO() << "[" << CS() << "];\n"
 		"\n"
-		"	_slen = " << SL() << "[" << CS() << "];\n"
-		"	if ( _slen > 0 ) {\n"
-		"		" << WIDE_ALPH_TYPE() << POINTER() << "_lower = _keys;\n"
-		"		" << WIDE_ALPH_TYPE() << POINTER() << "_mid;\n"
-		"		" << WIDE_ALPH_TYPE() << POINTER() << "_upper = _keys + _slen - 1;\n"
+		"	_klen = " << SL() << "[" << CS() << "];\n"
+		"	if ( _klen > 0 ) {\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_lower = _keys;\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_mid;\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_upper = _keys + _klen - 1;\n"
 		"		while (1) {\n"
 		"			if ( _upper < _lower )\n"
 		"				break;\n"
@@ -503,15 +586,15 @@ void TabCodeGen::LOCATE_TRANS()
 		"				goto _match;\n"
 		"			}\n"
 		"		}\n"
-		"		_keys += _slen;\n"
-		"		_trans += _slen;\n"
+		"		_keys += _klen;\n"
+		"		_trans += _klen;\n"
 		"	}\n"
 		"\n"
-		"	_rlen = " << RL() << "[" << CS() << "];\n"
-		"	if ( _rlen > 0 ) {\n"
-		"		" << WIDE_ALPH_TYPE() << POINTER() << "_lower = _keys;\n"
-		"		" << WIDE_ALPH_TYPE() << POINTER() << "_mid;\n"
-		"		" << WIDE_ALPH_TYPE() << POINTER() << "_upper = _keys + (_rlen<<1) - 2;\n"
+		"	_klen = " << RL() << "[" << CS() << "];\n"
+		"	if ( _klen > 0 ) {\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_lower = _keys;\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_mid;\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_upper = _keys + (_klen<<1) - 2;\n"
 		"		while (1) {\n"
 		"			if ( _upper < _lower )\n"
 		"				break;\n"
@@ -526,7 +609,7 @@ void TabCodeGen::LOCATE_TRANS()
 		"				goto _match;\n"
 		"			}\n"
 		"		}\n"
-		"		_trans += _rlen;\n"
+		"		_trans += _klen;\n"
 		"	}\n"
 		"\n";
 }
@@ -585,6 +668,7 @@ void TabCodeGen::RET( ostream &ret, bool inFinish )
 
 void TabCodeGen::BREAK( ostream &ret, int targState )
 {
+	outLabelUsed = true;
 	ret << "goto _out;";
 }
 
@@ -599,6 +683,28 @@ void TabCodeGen::writeOutData()
 		"\n";
 	}
 
+	if ( anyConditions() ) {
+		OPEN_ARRAY( ARRAY_TYPE(maxCondOffset), CO() );
+		COND_OFFSETS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( ARRAY_TYPE(maxCondLen), CL() );
+		COND_LENS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( WIDE_ALPH_TYPE(), CK() );
+		COND_KEYS();
+		CLOSE_ARRAY() <<
+		"\n";
+
+		OPEN_ARRAY( ARRAY_TYPE(maxCondSpaceId), C() );
+		COND_SPACES();
+		CLOSE_ARRAY() <<
+		"\n";
+	}
+
 	OPEN_ARRAY( ARRAY_TYPE(maxKeyOffset), KO() );
 	KEY_OFFSETS();
 	CLOSE_ARRAY() <<
@@ -677,28 +783,76 @@ void TabCodeGen::writeOutData()
 		"\n";
 	}
 
-	out <<
-		"static int " << START() << " = " << START_STATE_ID() << ";\n"
-		"\n";
+	STATIC_VAR( "int", START() ) << " = " << START_STATE_ID() << ";\n"
+	"\n";
 
 	if ( cgd->writeFirstFinal ) {
-		out <<
-			"static int " << FIRST_FINAL() << " = " << FIRST_FINAL_STATE() << ";\n"
-			"\n";
+		STATIC_VAR( "int" , FIRST_FINAL() ) << " = " << FIRST_FINAL_STATE() << ";\n"
+		"\n";
 	}
 
 	if ( cgd->writeErr ) {
-		out <<
-			"static int " << ERROR() << " = " << ERROR_STATE() << ";\n"
-			"\n";
+		STATIC_VAR( "int", ERROR() ) << " = " << ERROR_STATE() << ";\n"
+		"\n";
 	}
 }
 
+void TabCodeGen::COND_TRANSLATE()
+{
+	out << 
+		"	_widec = " << GET_KEY() << ";\n"
+		"	_klen = " << CL() << "[" << CS() << "];\n"
+		"	_keys = " << ARR_OFF( CK(), "(" + CO() + "[" + CS() + "]*2)" ) << ";\n"
+		"	if ( _klen > 0 ) {\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_lower = _keys;\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_mid;\n"
+		"		" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_upper = _keys + (_klen<<1) - 2;\n"
+		"		while (1) {\n"
+		"			if ( _upper < _lower )\n"
+		"				break;\n"
+		"\n"
+		"			_mid = _lower + (((_upper-_lower) >> 1) & ~1);\n"
+		"			if ( " << GET_WIDE_KEY() << " < _mid[0] )\n"
+		"				_upper = _mid - 2;\n"
+		"			else if ( " << GET_WIDE_KEY() << " > _mid[1] )\n"
+		"				_lower = _mid + 2;\n"
+		"			else {\n"
+		"				switch ( " << C() << "[" << CO() << "[" << CS() << "]"
+							" + ((_mid - _keys)>>1)] ) {\n";
+
+	for ( CondSpaceList::Iter csi = cgd->condSpaceList; csi.lte(); csi++ ) {
+		CondSpace *condSpace = csi;
+		out << "	case " << condSpace->condSpaceId << ": {\n";
+		out << TABS(2) << "_widec = " << KEY(condSpace->baseKey) << 
+				" + (" << GET_KEY() << " - " << KEY(keyOps->minKey) << ");\n";
+
+		for ( CondSet::Iter csi = condSpace->condSet; csi.lte(); csi++ ) {
+			out << TABS(2) << "if ( ";
+			CONDITION( out, *csi );
+			Size condValOffset = ((1 << csi.pos()) * keyOps->alphSize());
+			out << " ) _widec += " << condValOffset << ";\n";
+		}
+
+		out << 
+			"		break;\n"
+			"	}\n";
+	}
+	out << 
+		"				}\n"
+		"				break;\n"
+		"			}\n"
+		"		}\n"
+		"	}\n"
+		"\n";
+}
+
 void TabCodeGen::writeOutExec()
 {
+	outLabelUsed = false;
+
 	out <<
 		"	{\n"
-		"	int _slen, _rlen";
+		"	int _klen";
 
 	if ( anyRegCurStateRef() )
 		out << ", _ps";
@@ -710,48 +864,31 @@ void TabCodeGen::writeOutExec()
 	if ( anyConditions() )
 		out << "	" << WIDE_ALPH_TYPE() << " _widec;\n";
 
-	if ( anyActions() ) {
+	if ( anyToStateActions() || anyRegActions() || anyFromStateActions() ) {
 		out << 
-			"	" << ARRAY_TYPE(maxActArrItem) << " *_acts;\n"
+			"	" << PTR_CONST() << ARRAY_TYPE(maxActArrItem) << POINTER() << "_acts;\n"
 			"	" << UINT() << " _nacts;\n";
 	}
 
 	out <<
-		"	" << WIDE_ALPH_TYPE() << " *_keys;\n"
+		"	" << PTR_CONST() << WIDE_ALPH_TYPE() << POINTER() << "_keys;\n"
 		"\n";
 
-	out << 
-		"	" << P() << " -= 1;\n"
-		"	goto _resume;\n"
-		"\n";
-
-	/* Head of the loop. */
-	out << "_again:\n";
-
-	if ( anyToStateActions() ) {
-		out <<
-			"	_acts = " << ARR_OFF( A(), TSA() + "[" + CS() + "]" ) << ";\n"
-			"	_nacts = " << CAST(UINT()) << " *_acts++;\n"
-			"	while ( _nacts-- > 0 ) {\n"
-			"		switch ( *_acts++ ) {\n";
-			TO_STATE_ACTION_SWITCH();
-			SWITCH_DEFAULT() <<
-			"		}\n"
-			"	}\n"
-			"\n";
+	if ( cgd->hasEnd ) {
+		outLabelUsed = true;
+		out << 
+			"	if ( " << P() << " == " << PE() << " )\n"
+			"		goto _out;\n";
 	}
 
 	out << "_resume:\n";
 
-	if ( cgd->hasEnd ) {
+	if ( redFsm->errState != 0 ) {
+		outLabelUsed = true;
 		out << 
-			"	if ( ++" << P() << " == " << PE() << " )\n"
+			"	if ( " << CS() << " == " << redFsm->errState->id << " )\n"
 			"		goto _out;\n";
 	}
-	else {
-		out << 
-			"	" << P() << " += 1;\n";
-	}
 
 	if ( anyFromStateActions() ) {
 		out <<
@@ -771,12 +908,6 @@ void TabCodeGen::writeOutExec()
 
 	LOCATE_TRANS();
 
-	if ( redFsm->errState != 0 ) {
-		out << 
-			"	if ( " << CS() << " == " << redFsm->errState->id << " )\n"
-			"		goto _out;\n";
-	}
-
 	out << "_match:\n";
 
 	if ( anyRegCurStateRef() )
@@ -805,9 +936,34 @@ void TabCodeGen::writeOutExec()
 			"\n";
 	}
 
-	out << "	goto _again;\n";
+	if ( anyRegActions() || anyActionGotos() || anyActionCalls() || anyActionRets() )
+		out << "_again:\n";
+
+	if ( anyToStateActions() ) {
+		out <<
+			"	_acts = " << ARR_OFF( A(), TSA() + "[" + CS() + "]" ) << ";\n"
+			"	_nacts = " << CAST(UINT()) << " *_acts++;\n"
+			"	while ( _nacts-- > 0 ) {\n"
+			"		switch ( *_acts++ ) {\n";
+			TO_STATE_ACTION_SWITCH();
+			SWITCH_DEFAULT() <<
+			"		}\n"
+			"	}\n"
+			"\n";
+	}
+
+	if ( cgd->hasEnd ) {
+		out << 
+			"	if ( ++" << P() << " != " << PE() << " )\n"
+			"		goto _resume;\n";
+	}
+	else {
+		out << 
+			"	" << P() << " += 1;\n"
+			"	goto _resume;\n";
+	}
 	
-	if ( cgd->hasEnd || anyRegBreak() || redFsm->errState != 0 )
+	if ( outLabelUsed )
 		out << "	_out: {}\n";
 
 	out << "	}\n";
@@ -819,7 +975,7 @@ void TabCodeGen::writeOutEOF()
 	if ( anyEofActions() ) {
 		out << 
 			"	{\n"
-			"	" << ARRAY_TYPE(maxActArrItem) << " *_acts = " << 
+			"	" << PTR_CONST() << ARRAY_TYPE(maxActArrItem) << POINTER() << "_acts = " << 
 					ARR_OFF( A(), EA() + "[" + CS() + "]" ) << ";\n"
 			"	" << UINT() << " _nacts = " << CAST(UINT()) << " *_acts++;\n"
 			"	while ( _nacts-- > 0 ) {\n"
